from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.index import index
class index_unique(parser, index):

    _index_type = 'unique'
    has_comma = False

    # UNIQUE account_id (account_id)
    rules = [{
        'type': 'literal',
        'value': 'UNIQUE'
    }, {
        'type': 'regexp',
        'name': 'index_name',
        'value': '(KEY)|(INDEX)'
    }, {
        'type': 'regexp',
        'name': 'name',
        'value': '[^\(]+'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'delimited',
        'name': 'columns',
        'separator': ',',
        'quote': '`'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def __init__(self, rules=[]):

        super().__init__(rules)

        self._errors = []
        self._warnings = []
        self._columns = []

    def process(self):

        self._name = self._values['name'].strip().strip('`')
        self._columns = self._values['columns']
        self.has_comma = True if 'ending_comma' in self._values else False

        if len(self._name) > 64:
            self._errors.append('Key name %s is too long' % (self._name))



import unittest

from mygrations.formats.mysql.file_reader.reader import reader
class test_reader(unittest.TestCase):
    def test_simple(self):

        parser = reader()
        returned = parser.parse(
            """
            /* I should find and ignore comments */
            -- Of any variety
               # really

            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;

            INSERT INTO logs (`id`,`message`,`traceback`) VALUES (1,'blah','gah');
            INSERT INTO test (id,status) VALUES (1,'New'),(2,'Old');
        """
        )

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have matched everything
        self.assertEquals('', returned)

        # our parser should have a table!
        self.assertTrue('logs' in parser.tables)
        # just check a few details
        self.assertEquals('logs', parser.tables['logs'].name)
        self.assertEquals(3, len(parser.tables['logs'].columns))

        # and some records!
        self.assertTrue('logs' in parser.rows)
        self.assertTrue('test' in parser.rows)
        self.assertEquals(['id', 'message', 'traceback'], parser.rows['logs'][0].columns)



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
class test_table_decimal_false_positives(unittest.TestCase):
    def test_decimal_false_positives(self):
        """ false positives for change on default for decimal columns

        The default value for float and decimal columns can have the same issue for
        equality checking as any other floating comparison.  If the user's file sets
        the default to `0` mysql will return a default of '0.00', leading to a false
        detection of a column change and unnecessary database changes.
        """
        from_table = create_parser()
        from_table.parse(
            """CREATE TABLE `decimal_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `price` decimal(20,2) NOT NULL DEFAULT '0.00',
  PRIMARY KEY (`id`),
) ENGINE=InnoDB;
        """
        )

        to_table = create_parser()
        to_table.parse(
            """CREATE TABLE `decimal_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `price` decimal(20,2) NOT NULL DEFAULT 0,
  PRIMARY KEY (`id`),
) ENGINE=InnoDB;
        """
        )

        self.assertEquals(0, len(from_table.to(to_table)))

    def test_decimal_false_positives_just_because(self):
        """ This probably isn't a realistic test, but let's go for it while we're here """
        from_table = create_parser()
        from_table.parse(
            """CREATE TABLE `decimal_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `price` decimal(20,2) NOT NULL DEFAULT '1.007',
  PRIMARY KEY (`id`),
) ENGINE=InnoDB;
        """
        )

        to_table = create_parser()
        to_table.parse(
            """CREATE TABLE `decimal_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `price` decimal(20,2) NOT NULL DEFAULT '1.008',
  PRIMARY KEY (`id`),
) ENGINE=InnoDB;
        """
        )

        self.assertEquals(0, len(from_table.to(to_table)))

    def test_decimal_real_positives(self):
        from_table = create_parser()
        from_table.parse(
            """CREATE TABLE `decimal_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `price` decimal(20,2) NOT NULL DEFAULT '1.00',
  PRIMARY KEY (`id`),
) ENGINE=InnoDB;
        """
        )

        to_table = create_parser()
        to_table.parse(
            """CREATE TABLE `decimal_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `price` decimal(20,2) NOT NULL DEFAULT '1.01',
  PRIMARY KEY (`id`),
) ENGINE=InnoDB;
        """
        )

        self.assertEquals(1, len(from_table.to(to_table)))

    def test_float_false_positives(self):
        from_table = create_parser()
        from_table.parse(
            """CREATE TABLE `decimal_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `price` float NOT NULL DEFAULT '0.00',
  PRIMARY KEY (`id`),
) ENGINE=InnoDB;
        """
        )

        to_table = create_parser()
        to_table.parse(
            """CREATE TABLE `decimal_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `price` float NOT NULL DEFAULT 0,
  PRIMARY KEY (`id`),
) ENGINE=InnoDB;
        """
        )

        self.assertEquals(0, len(from_table.to(to_table)))



import copy

from mygrations.formats.mysql.definitions.database import database
from mygrations.formats.mysql.mygrations.operations.row_insert import row_insert
from mygrations.formats.mysql.mygrations.operations.row_delete import row_delete
from mygrations.formats.mysql.mygrations.operations.row_update import row_update
class row_mygration:
    """ Creates a migration plan to sync the rows of a database for tables being tracked.

    This doesn't sync rows for all tables but only those tables marked as
    tracking_rows
    """

    def __init__(self, db_to, db_from=None):
        """ Create a migration plan

        The row migrator assumes that both databases have already read rows into the
        table objects.  In standard usage, the files database will automatically
        read in the table rows (i.e. mygrations.formats.mysql.file_reader.database),
        but the actual database reader (mygrations.formats.mysql.db_reader.database)
        will not.  This is because, in normal usage, only the rows in *some* tables
        are tracked.  Which tables to track rows for is normally denoted by the fact
        that the *.sql files have inserts in them, which the file_reader picks up on.
        Then, the db_reader is told to read rows in only for those tables where the
        file_reader is tracking rows.  I.e.:

        >>> for table in files_database.tables.values():
        >>>     if not table.tracking_rows or table.name not in live_database.tables:
        >>>         continue
        >>>     live_database.read_rows( table )

        :param db_to: The target database rows to migrate to
        :param db_from: The current database rows to migrate from
        :type db_to: mygrations.formats.mysql.definitions.database
        :type db_from: mygrations.formats.mysql.definitions.database
        """
        self.db_to = db_to
        self.db_from = db_from

        # unlike with the structure migration, we figure out what are operations
        # are and then make sure those operations won't cause any 1215 errors
        self._operations = self._process()

        self._errors_1215 = []

    @property
    def operations(self):
        """ Public getter.  Returns list of operations to bring rows in db_from to match db_to

        If db_from doesn't exist then it will be a list of operations to
        create the rows in db_to.

        If the two databases have the exact same  rows for tables that are tracking_rows
        then operations will be an empty list

        :returns: A list of table row operations
        :rtype: None|[mygrations.formats.mysql.mygrations.operations.operation]
        """
        return self._operations

    @property
    def errors_1215(self):
        """ Public getter.  Returns list of 1215 errors (as strings)

        :returns: A list of 1215 error messages
        :rtype: [string]
        """
        return self._errors_1215

    def __len__(self):
        return len(self._operations)

    def __bool__(self):
        return True if len(self._operations) else False

    def __str__(self):
        return "\n".join([str(x) for x in self._operations])

    def __iter__(self):
        return self._operations.__iter__()

    def _process(self):
        """ Figures out the row operations needed to get to self.db_to

        Returns a list of operations that will migrate rows in db_from to db_to

        :return: Operations needed to complete the migration
        :rtype: [mygrations.formats.mysql.mygrations.operations]
        """
        operations = []
        tracking_tables = [table.name for table in self.db_to.tables.values() if table.tracking_rows]
        for table_name in tracking_tables:
            from_table = self.db_from.tables[table_name] if (
                self.db_from and table_name in self.db_from.tables
            ) else None
            more_operations = self.db_to.tables[table_name].to_rows(from_table)
            if not more_operations:
                continue

            operations.extend(more_operations)

        return operations



class change_column:
    """ Generates a partial SQL command to change a column in a table """

    def __init__(self, column):
        self.column = column

    def apply_to_table(self, table):
        """ Changes the column in the table

        :param table: The table to change the column on
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.change_column(self.column)

    def __str__(self):
        return 'CHANGE `%s` %s' % (self.column.name, str(self.column))



import unittest

from mygrations.formats.mysql.file_reader.parsers.index_key import index_key
class test_index_key(unittest.TestCase):
    def test_simple(self):

        # parse a typical KEY
        parser = index_key()
        returned = parser.parse('KEY `users_email` (`email`),')

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals('', returned)

        # we should have lots of data now
        self.assertEquals('users_email', parser.name)
        self.assertEquals(['email'], parser.columns)
        self.assertTrue(parser.has_comma)
        self.assertEquals('KEY `users_email` (`email`)', str(parser))

    def test_optional_comma(self):

        # ending comma is optional
        parser = index_key()
        returned = parser.parse('KEY `users_email` (`email`)')

        # we should have matched
        self.assertTrue(parser.matched)
        self.assertFalse(parser.has_comma)
        self.assertEquals('KEY `users_email` (`email`)', str(parser))

    def test_optional_quotes(self):

        # key name quotes are optional
        parser = index_key()
        returned = parser.parse('KEY users_email (`email`)')

        # we should have matched
        self.assertTrue(parser.matched)
        self.assertEquals('users_email', parser.name)
        self.assertEquals('KEY `users_email` (`email`)', str(parser))

    def test_multiple_columns(self):

        # multi column index
        parser = index_key()
        returned = parser.parse('KEY `users_email` (`email`,`username`,`password`),')

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals('', returned)

        # we should have lots of data now
        self.assertEquals('users_email', parser.name)
        self.assertEquals(['email', 'username', 'password'], parser.columns)
        self.assertTrue(parser.has_comma)
        self.assertEquals('KEY `users_email` (`email`,`username`,`password`)', str(parser))



import unittest

from mygrations.core.parse.rule_regexp import rule_regexp
class test_parse_regexp(unittest.TestCase):
    def get_rule(self, name, regexp):

        return rule_regexp(False, {'name': name, 'value': regexp}, {})

    def test_name_required(self):

        with self.assertRaises(ValueError):

            self.get_rule('', '')

    def test_value_required(self):

        with self.assertRaises(ValueError):

            self.get_rule('bob', '')

    def test_can_init_with_name_and_value(self):

        rule = self.get_rule('bob', '\S+')
        self.assertEquals(rule.name, 'bob')
        self.assertEquals(rule.regexp, '\S+')

    def test_match_beginning_only(self):

        rule = self.get_rule('bob', '\d+')
        self.assertFalse(rule.parse('hey 123'))
        self.assertEquals('', rule.result)

    def test_leftovers_is_input_for_no_match(self):

        rule = self.get_rule('bob', '\d+')
        string = 'hey 123'
        rule.parse(string)

        self.assertEquals(string, rule.leftovers)

    def test_no_leftovers_for_full_match(self):

        rule = self.get_rule('bob', '\d+')
        string = '23483438'

        self.assertTrue(rule.parse(string))
        self.assertEquals(string, rule.result)
        self.assertEquals('', rule.leftovers)

    def test_return_group_only(self):

        rule = self.get_rule('bob', '\d+\s+(\w+)')

        string = '999 bob'

        self.assertTrue(rule.parse(string))
        self.assertEquals('bob', rule.result)
        self.assertEquals('', rule.leftovers)

    def test_calc_leftovers_trim(self):

        rule = self.get_rule('bob', '\d+')

        string = '999 bob'

        self.assertTrue(rule.parse(string))
        self.assertEquals('999', rule.result)
        self.assertEquals('bob', rule.leftovers)



import unittest

from mygrations.formats.mysql.file_reader.parsers.index_primary import index_primary
class test_index_primary(unittest.TestCase):
    def test_simple(self):

        # parse a typical KEY
        parser = index_primary()
        returned = parser.parse('PRIMARY KEY (`id`),')

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals('', returned)

        # we should have lots of data now
        self.assertEquals(['id'], parser.columns)
        self.assertEquals(parser.index_type, 'PRIMARY')
        self.assertTrue(parser.has_comma)
        self.assertEquals('PRIMARY KEY (`id`)', str(parser))

    def test_optional_comma(self):

        # ending comma is optional
        parser = index_primary()
        returned = parser.parse('PRIMARY KEY (`id`)')

        # we should have matched
        self.assertEquals(['id'], parser.columns)
        self.assertTrue(parser.matched)
        self.assertFalse(parser.has_comma)
        self.assertEquals('PRIMARY KEY (`id`)', str(parser))

    def test_optional_quotes(self):

        # key name quotes are optional
        parser = index_primary()
        returned = parser.parse('PRIMARY KEY (sup)')

        # we should have matched
        self.assertTrue(parser.matched)
        self.assertEquals(['sup'], parser.columns)
        self.assertEquals('PRIMARY KEY (`sup`)', str(parser))



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
class test_table_difference(unittest.TestCase):
    def _get_parsers(self):
        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id)
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id),
            KEY tasks_account_id (account_id),
            KEY task (task,account_id)
            );
        """
        )

        return (a, b)

    def test_drop_keys(self):
        (a, b) = self._get_parsers()

        # if we subtract b from a we should get some drop column queries in one alter statement
        operations = b.to(a)
        self.assertEquals(1, len(operations))
        self.assertEquals('ALTER TABLE `tasks` DROP KEY `tasks_account_id`, DROP KEY `task`;', str(operations[0]))

    def test_add_keys(self):
        (a, b) = self._get_parsers()

        # if we subtract b from a we should get some drop column queries in one alter statement
        operations = a.to(b)
        self.assertEquals(1, len(operations))
        self.assertEquals(
            'ALTER TABLE `tasks` ADD KEY `tasks_account_id` (`account_id`), ADD KEY `task` (`task`,`account_id`);',
            str(operations[0])
        )

    def test_change_keys(self):
        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id),
            KEY tasks_account_id (account_id)
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id),
            KEY tasks_account_id (account_id,id,task)
            );
        """
        )

        # if we subtract b from a we should get some drop column queries in one alter statement
        operations = a.to(b)
        self.assertEquals(1, len(operations))
        self.assertEquals(
            'ALTER TABLE `tasks` DROP KEY `tasks_account_id`, ADD KEY `tasks_account_id` (`account_id`,`id`,`task`);',
            str(operations[0])
        )

    def test_add_remove_change(self):
        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id),
            KEY unchanged (task),
            KEY tasks_task (task),
            KEY tasks_account_id (account_id)
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id),
            KEY unchanged (task),
            KEY task_id (task,id),
            KEY tasks_account_id (account_id,id,task)
            );
        """
        )

        # if we subtract b from a we should get some drop column queries in one alter statement
        operations = a.to(b)
        self.assertEquals(1, len(operations))
        self.assertEquals(
            'ALTER TABLE `tasks` ADD KEY `task_id` (`task`,`id`), DROP KEY `tasks_task`, DROP KEY `tasks_account_id`, ADD KEY `tasks_account_id` (`account_id`,`id`,`task`);',
            str(operations[0])
        )

    def test_unique_keys(self):
        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id),
            UNIQUE KEY unchanged (account_id),
            UNIQUE KEY uniq_account_id (account_id),
            unique key uniq_task (task)
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id),
            UNIQUE KEY unchanged (account_id),
            UNIQUE KEY uniq_account_id (account_id,id),
            unique key added (task)
            );
        """
        )

        operations = a.to(b)
        self.assertEquals(1, len(operations))
        self.assertEquals(
            'ALTER TABLE `tasks` ADD UNIQUE KEY `added` (`task`), DROP KEY `uniq_task`, DROP KEY `uniq_account_id`, ADD UNIQUE KEY `uniq_account_id` (`account_id`,`id`);',
            str(operations[0])
        )

    def test_primary_key(self):
        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id)
            );
        """
        )

        operations = a.to(b)
        self.assertEquals(1, len(operations))
        self.assertEquals('ALTER TABLE `tasks` ADD PRIMARY KEY (`id`);', str(operations[0]))

        operations = b.to(a)
        self.assertEquals(1, len(operations))
        self.assertEquals('ALTER TABLE `tasks` DROP PRIMARY KEY;', str(operations[0]))



import unittest

from mygrations.formats.mysql.file_reader.insert_parser import insert_parser
class test_insert_parser(unittest.TestCase):
    def test_simple(self):

        parser = insert_parser()
        returned = parser.parse("INSERT INTO test_table (`col1`,`col2`) VALUES ('val','val2');")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have matched everything
        self.assertEquals('', returned)

        # we should have lots of data now
        self.assertEquals('test_table', parser.table)
        self.assertEquals(['col1', 'col2'], parser.columns)
        self.assertEquals([['val', 'val2']], parser.raw_rows)
        self.assertTrue(parser.has_semicolon)
        self.assertEquals([], parser.errors)

    def test_multiple_values(self):

        parser = insert_parser()
        returned = parser.parse("INSERT INTO test_table (`col1`,`col2`) VALUES ('val','val2'),('val3','val4')")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have matched everything
        self.assertEquals('', returned)

        # we should have lots of data now
        self.assertEquals('test_table', parser.table)
        self.assertEquals(['col1', 'col2'], parser.columns)
        self.assertEquals([['val', 'val2'], ['val3', 'val4']], parser.raw_rows)
        self.assertFalse(parser.has_semicolon)
        self.assertEquals([], parser.errors)

    def test_missing_comma(self):

        parser = insert_parser()
        returned = parser.parse("INSERT INTO test_table (`col1`,`col2`) VALUES ('val','val2')('val3','val4')")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have matched everything
        self.assertEquals('', returned)

        # we should have lots of data now
        self.assertEquals('test_table', parser.table)
        self.assertEquals(['col1', 'col2'], parser.columns)
        self.assertEquals([['val', 'val2'], ['val3', 'val4']], parser.raw_rows)
        self.assertFalse(parser.has_semicolon)
        self.assertEquals(1, len(parser.warnings))



import MySQLdb

from .base import base
from mygrations.formats.mysql.file_reader.database import database as database_parser
from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.formats.mysql.mygrations.mygration import mygration
def execute(options):

    obj = check(options)
    obj.execute()
class check(base):
    def execute(self):
        files_database = database_parser(self.config['files_directory'])

        # any errors or warnings?
        errors = False
        if files_database.errors:
            print('Errors found in *.sql files')
            errors = True
            for error in files_database.errors:
                print(error)

        errors_1215 = files_database.errors_1215
        if errors_1215:
            print('1215 Errors encountered')
            errors = True
            for error in errors_1215:
                print(error)

        if not errors:
            print("No problems found")



from collections import OrderedDict

from .base import base
from mygrations.formats.mysql.file_reader.database import database as database_parser
from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.formats.mysql.mygrations.mygration import mygration
from mygrations.formats.mysql.mygrations.row_mygration import row_mygration
from mygrations.formats.mysql.mygrations.operations.row_insert import row_insert
from mygrations.drivers.mysqldb.mysqldb import mysqldb
def execute(options):

    obj = plan_export(options)
    obj.execute()
class plan_export(base):
    def execute(self):

        files_database = database_parser(self.config['files_directory'])

        # use the credentials to load up a database connection
        live_database = database_reader(mysqldb(self.credentials))

        # we aren't outputting operations.  Instead we just need to know what tables
        # have changed (either structure or records).  The easiest (and slightly hack)
        # way to do that is to simply run an actual mygration and grab out the names
        # of the tables that have changed.  Cheating, I know
        self.modified_tables = {}
        mygrate = mygration(live_database, files_database, False)
        if mygrate.operations:
            for op in mygrate.operations:
                self.modified_tables[op.table_name] = True

        # we have to tell the live database to load records
        # for any tables we are tracking records for, according
        # to the files database.
        for table in files_database.tables.values():
            if not table.tracking_rows:
                continue
            if not table.name in live_database.tables:
                continue
            live_database.read_rows(table)

        rows = row_mygration(live_database, files_database)
        if rows.operations:
            for op in rows.operations:
                self.modified_tables[op.table_name] = True

        for table_name in self.modified_tables.keys():
            print('\033[1;33m\033[41m%s\033[0m' % table_name)
            table = live_database.tables[table_name]
            print(table.nice())

            if table.tracking_rows:
                print('\n')
                for row in table.rows.values():
                    print(row_insert(table.name, row))



class row_insert:
    """ Generates an SQL command to insert a record """

    def __init__(self, table_name, data):
        if type(table_name) != str:
            self._table_name = table_name.name
        else:
            self._table_name = table_name
        self.data = data

    @property
    def table_name(self):
        """ Public getter.  Returns the name of the table.

        :returns: The table name
        :rtype: string
        """

        return self._table_name

    def __str__(self):
        cols = ', '.join(['`%s`' % val for val in self.data.keys()])
        vals = ', '.join([
            "'%s'" % str(val).replace('\\', '\\\\').replace("'", "\\'") if val is not None else 'NULL'
            for val in self.data.values()
        ])
        return 'INSERT INTO `%s` (%s) VALUES (%s);' % (self._table_name, cols, vals)



from .rule_base import rule_base
class rule_children(rule_base):
    """ rule = rule_children( parser, rule, next_rule )

    rule_children constructor.  Pass in the parser object that the rule is for, the rule
    that is being parsed, and the next rule in the parsing chain.

    This class takes as input a list of parser classes to be matched against the string
    being parsed.  During parsing, it will attempt to fit any of the child parsers (in
    any order) against the string being parsed.  After each match, the matched contents
    will be removed and parsing will start again.  This repeats until no more matches
    can be found.

    keys for the rules dictionary:

    =========  ========
    key        contents
    =========  ========
    name       The name of the rule
    classes    A list with classes to be matched
    optional   (optional) If true, denotes that this rule is optional
    =========  ========

    :param parser: The parser that this rule belongs to
    :param rule: A dictionary representing the rule configuration (see each rule class for details)
    :param next_rule: A dictionary representing the next rule
    :type parser: parser
    :type rule: dict
    :type next_rule: dict
    :returns: rule object
    """

    require_value = False

    classes = []

    def __init__(self, parser, rule, next_rule):

        super().__init__(parser, rule, next_rule)

        if not 'classes' in self.rule:
            raise ValueError('Missing classes for children rule %s in %s' % (rule, self.parser_class))

        if not self.rule['classes']:
            raise ValueError('Missing classes for children rule %s in %s' % (rule, self.parser_class))

        self.classes = self.rule['classes']

    def parse(self, string):

        matches = []

        # we need to keep looping through the children looking for matches
        # to the string until we no longer get any matches.  Then return.
        c = 0
        while string:

            c += 1
            if c > 1000:
                raise ValueError('Max depth reached')

            shortest_leftover = False
            best_match = False
            best_leftover = ''
            current_string = string

            for child in self.classes:

                # let the child parse the string
                child_parser = child()
                leftover = child_parser.parse(current_string)

                # if it didn't do anything then this child didn't match
                if not child_parser.matched:
                    continue

                # It is possible that more than one child matches.  If so,
                # prioritize the child that matches the largest part of the
                # string.  All else being equal, first come first serve
                if best_match:
                    if len(leftover) < shortest_leftover:
                        shortest_leftover = len(leftover)
                        best_match = child_parser
                        best_leftover = leftover
                else:
                    shortest_leftover = len(leftover)
                    best_match = child_parser
                    best_leftover = leftover

            # we have a match!
            if best_match:
                string = best_leftover
                matches.append(best_match)

            # If we didn't find anything then we are completely done
            else:
                if matches:
                    self.leftovers = string
                    self.result = matches
                    return True
                else:
                    self.leftovers = string
                    self.result = ''
                return False

        # the only way we would get here is if we matched the entire string
        self.leftovers = ''
        self.result = matches
        return True



class add_column:
    """ Generates a partial SQL command to add a column to a table """

    def __init__(self, column, position=None):
        """ Create an add column operation

        Position should be one of:

        ==================  ====================
        Value               Meaning
        ==================  ====================
        None                Put at the end of the table
        False               Put at the end of the table
        True                Put at the beginning of the table
        Any string          position should be a column name: put the new column after that one
        ==================  ====================

        :param column: The column to add
        :param position: Where to put the new column in the table
        :type column: mygrations.formats.mysql.definitions.column
        :type position: mixed:
        """

        self.column = column
        self.position = position

    def apply_to_table(self, table):
        """ Adds the column to the table

        :param table: The table to add the column to
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.add_column(self.column, self.position)

    def __str__(self):
        basic = 'ADD %s' % self.column
        if self.position == True:
            position = ' FIRST'
        elif type(self.position) == type(''):
            position = ' AFTER `%s`' % self.position
        else:
            position = ''

        return 'ADD %s%s' % (str(self.column), position)



import unittest

from mygrations.core.parse.rule_children import rule_children
from mygrations.core.parse.parser import parser
class test_rule(parser):

    rules = [{'type': 'literal', 'value': 'bob'}, {'type': 'literal', 'value': 'greg'}]
class test_rule_two(parser):

    rules = [{'type': 'regexp', 'name': 'test_regexp', 'value': '\d+'}, {'type': 'literal', 'value': 'sup'}]
class test_rule_overlap(parser):

    rules = [{'type': 'regexp', 'name': 'test_overlap', 'value': '\S+'}]
class test_parse_children(unittest.TestCase):
    def get_rule(self, name, classes):

        return rule_children(False, {'name': name, 'classes': classes}, {})

    def test_name_required(self):

        with self.assertRaises(ValueError):

            self.get_rule('', [test_rule])

    def test_classes_required(self):

        with self.assertRaises(ValueError):

            self.get_rule('bob', [])

    # if nothing matches at the start then we get no match
    def test_match_start_only(self):

        rule = self.get_rule('test', [test_rule, test_rule_two])
        self.assertFalse(rule.parse('okay bob greg'))
        self.assertEquals('', rule.result)
        self.assertEquals('okay bob greg', rule.leftovers)

    # we should be able to match any number of children in any order
    # as long as their parsers match, we will match them
    def test_match_any_order(self):

        rule = self.get_rule('test', [test_rule, test_rule_two])
        self.assertTrue(rule.parse('999 sup bob greg bob greg 1234568 sup hey'))
        self.assertEquals('hey', rule.leftovers)

        self.assertEquals(4, len(rule.result))
        self.assertTrue(rule.result[0].__class__ == test_rule_two)
        self.assertTrue(rule.result[1].__class__ == test_rule)
        self.assertTrue(rule.result[2].__class__ == test_rule)
        self.assertTrue(rule.result[3].__class__ == test_rule_two)

    # if two rules both match a string the rule that matches
    # the largest portion of the string wins.
    def test_matches_most_wins(self):

        rule = self.get_rule('test', [test_rule_overlap, test_rule])

        self.assertTrue(rule.parse('bob greg'))
        self.assertTrue(rule.result[0].__class__ == test_rule)
        self.assertEquals('', rule.leftovers)

    # if a child can't match all required params other
    # children can match
    def test_missing_required_cant_match(self):

        rule = self.get_rule('test', [test_rule_overlap, test_rule])

        self.assertTrue(rule.parse('bob'))
        self.assertTrue(rule.result[0].__class__ == test_rule_overlap)
        self.assertEquals('', rule.leftovers)



import unittest
from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.formats.mysql.file_reader.database import database as file_reader
from mygrations.drivers.mysqldb.mysqldb import mysqldb
from tests.mocks.db.mysql.db_structure import db_structure
from mygrations.formats.mysql.mygrations.row_mygration import row_mygration
class test_mygrate_null(unittest.TestCase):
    def test_diff_with_null(self):
        """ NULL should be allowed and should result in a MySQL NULL in the database

        The system was turning NULL into a literal 'NULL'.  Internally, NULL
        is handled as a None
        """
        tables = {
            'logs':
            """
                CREATE TABLE `logs` (
                    `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                    `message` TEXT NOT NULL,
                    `traceback` text,
                    PRIMARY KEY (`id`)
                ) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        }

        rows = {
            'logs': (
                {
                    'id': 1,
                    'message': 'from null to value',
                    'traceback': None
                },
                {
                    'id': 2,
                    'message': 'from value to null',
                    'traceback': 'forever'
                },
                {
                    'id': 3,
                    'message': 'from null to null',
                    'traceback': None
                },
            )
        }

        db_db = database_reader(mysqldb(db_structure(tables, rows)))
        db_db.read_rows('logs')

        # and one from a file
        table1 = """CREATE TABLE `logs` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `message` TEXT NOT NULL,
            `traceback` text,
            PRIMARY KEY (`id`)
        ) ENGINE=InnoDB DEFAULT CHARSET=utf8;

        INSERT INTO logs (id,message,traceback) VALUES (1,'from null to value', 'HEY');
        INSERT INTO logs (id,message,traceback) VALUES (2,'from value to null', NULL);
        INSERT INTO logs (id,message,traceback) VALUES (3,'from null to null', NULL);
        INSERT INTO logs (id,message,traceback) VALUES (4,'Insert Null',NULL);
        """
        files_db = file_reader([table1])

        mygrate = row_mygration(files_db, db_db)
        ops = [str(op) for op in mygrate]
        self.assertEquals(3, len(ops))
        self.assertTrue("INSERT INTO `logs` (`id`, `message`, `traceback`) VALUES ('4', 'Insert Null', NULL);" in ops)
        self.assertTrue("UPDATE `logs` SET `message`='from null to value', `traceback`='HEY' WHERE id=1;" in ops)
        self.assertTrue("UPDATE `logs` SET `message`='from value to null', `traceback`=NULL WHERE id=2;" in ops)



import unittest

from mygrations.helpers.dotenv import dotenv, DotEnvSyntaxError
# mainly just breakning the tests up a little bit.
# the ability to recognize and parse around quotes inside
# the values does make this a bit more complicated, so it
# makes sense to be in another file anyway
class test_dotenv_parse_line_with_quotes(unittest.TestCase):

    dotenv = None

    def setUp(self):

        self.dotenv = dotenv()

    def test_no_lone_quotes_double(self):

        # a quote character inside the value by itself is invalid
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, 'name = valu"e', '#')

    def test_no_lone_quotes_single(self):

        # a quote character inside the value by itself is invalid
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, "name = valu'e", '#')

    def test_empty_single(self):

        # easy
        self.assertEquals(['db', ''], self.dotenv.parse_line("db = ''", '#'))

    def test_empty_double(self):

        # easy
        self.assertEquals(['db', ''], self.dotenv.parse_line('db = ""', '#'))

    def test_empty_single_comment(self):

        # easy
        self.assertEquals(['db', ''], self.dotenv.parse_line("db = '' # database name", '#'))

    def test_empty_double_comment(self):

        # easy
        self.assertEquals(['db', ''], self.dotenv.parse_line('db = "" ; database name', ';'))

    def test_single_quotes(self):

        # I shouldn't get the quotes back.  Also white space should
        # be ignored still
        self.assertEquals(['db', 'dijere'], self.dotenv.parse_line("db = 'dijere' ", '#'))

    def test_double_quotes(self):

        # I shouldn't get the quotes back.  Also white space should
        # be ignored still
        self.assertEquals(['db', 'dijere'], self.dotenv.parse_line('db = "dijere" ', '#'))

    def test_no_closing_quote_single(self):

        # syntax error if we have an opening quote with no close
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, "name = 'test", '#')

    def test_no_closing_quote_double(self):

        # syntax error if we have an opening quote with no close
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, 'name = "test', '#')

    def test_double_quotes_with_comment(self):

        # comments after the quotes should be ignored (along with whitespace)
        self.assertEquals(['db', 'bob'], self.dotenv.parse_line('db = "bob" ; database name', ';'))

    def test_single_quotes_with_comment(self):

        # comments after the quotes should be ignored (along with whitespace)
        self.assertEquals(['db', 'bob'], self.dotenv.parse_line("db = 'bob' \t# database name", '#'))

    def test_text_outside_of_double(self):

        # anything outside of the quotes and before a comment results in a syntax error
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, 'name = "test" sup # hey', '#')

    def test_text_outside_of_single(self):

        # anything outside of the quotes and before a comment results in a syntax error
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, "name = 'test' sup ; hey", ';')

    def test_allow_ending_semicolon(self):

        # anything outside of the quotes and before a comment results in a syntax error
        # except allow to end the line with a semi-colon.
        self.assertEquals(['name', 'test'], self.dotenv.parse_line("name = 'test';", '#'))

    def test_escaped_single_quote(self):

        # quotes can be escaped inside quotes
        self.assertEquals(['db', "asdf'qwerty'"], self.dotenv.parse_line("db = 'asdf\\'qwerty\\''", '#'))

    def test_escaped_double_quote(self):

        # quotes can be escaped inside quotes
        self.assertEquals(['db', 'asdf"qwerty"'], self.dotenv.parse_line('db = "asdf\\"qwerty\\""', '#'))

    def test_preserve_other_slashes(self):

        # other slashes are left alone
        self.assertEquals(['db', 'asdf\\bob'], self.dotenv.parse_line('db = "asdf\\bob"', '#'))

    def test_double_quote_in_single_quote(self):

        # double quote inside single quotes are just regular characters
        self.assertEquals(['db', 'asdf"bob'], self.dotenv.parse_line("db = 'asdf\"bob'", '#'))

    def test_single_quote_in_double_quote(self):

        # single quote inside double quotes are just regular characters
        self.assertEquals(['db', "asdf'bob"], self.dotenv.parse_line('db = "asdf\'bob"', '#'))



class alter_table:
    """ Generates an SQL command to alter a table """

    def __init__(self, table):
        self._table = table
        self._operations = []

    def add_operation(self, operation):
        self._operations.append(operation)

    @property
    def table_name(self):
        """ Public getter.  Returns the name of the table.

        :returns: The table name
        :rtype: string
        """

        return self._table

    def __len__(self):
        return len(self._operations)

    def __bool__(self):
        return True if len(self._operations) else False

    def __str__(self):
        return 'ALTER TABLE `%s` %s;' % (self._table, ', '.join([str(x) for x in self._operations]))

    def __iter__(self):
        return self._operations.__iter__()



from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.column import column
class type_text(parser, column):

    #***** Lint error if default is found for text field ******
    allowed_types = {
        'tinyblob': True,
        'blob': True,
        'mediumblob': True,
        'longblob': True,
        'tinytext': True,
        'text': True,
        'mediumtext': True,
        'longtext': True
    }

    has_comma = False

    # description text charset utf8 collate utf8
    # hackish? maybe (i.e. yes) the repeated COLLATE and CHARACTER SETS take care of uncertain
    # ordering.  I could also use `children` which is order-agnostic, but I'm being lazy
    rules = [{
        'type': 'regexp',
        'value': '[^\(\s\)]+',
        'name': 'name'
    }, {
        'type': 'regexp',
        'value': '\w+',
        'name': 'type'
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'CHARACTER SET ([^\(\s\),]+)',
        'name': 'character_set',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'NOT NULL',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'CHARACTER SET ([^\(\s\),]+)',
        'name': 'character_set',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'DEFAULT ([^\(\s\),]+)',
        'optional': True,
        'name': 'default'
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'CHARACTER SET ([^\(\s\),]+)',
        'name': 'character_set',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):

        self.has_comma = True if 'ending_comma' in self._values else False

        self._errors = []
        self._warnings = []
        self._name = self._values['name'].strip('`')
        self._length = ''
        self._default = None
        self._column_type = self._values['type']
        self._null = False if 'NOT NULL' in self._values else True
        self._character_set = self._values['character_set'].strip("'") if 'character_set' in self._values else ''
        self._collate = self._values['collate'].strip("'") if 'collate' in self._values else ''

        if self._column_type.lower() == 'datetime':
            pass
        elif not self._column_type.lower() in self.allowed_types:
            self._errors.append('Column of type %s must have a length for column %s' % (self._column_type, self._name))
        elif 'default' in self._values:
            self._errors.append(
                'Column of type %s is not allowed to have a default value for column %s' %
                (self._column_type, self._name)
            )



import unittest

from mygrations.formats.mysql.file_reader.database import database as database_reader
from mygrations.formats.mysql.file_reader.create_parser import create_parser
from mygrations.formats.mysql.mygrations.mygration import mygration
class test_add_only(unittest.TestCase):
    def test_add_one_only(self):
        """ mygration can accept only one table, in which case it just figures out the proper add order

        If there are no foreign key constraints, then it will "add" them in random order
        """

        table1 = "CREATE TABLE `zlogs` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,\n`message` TEXT NOT NULL,\n`traceback` TEXT,\nPRIMARY KEY (`id`)) ENGINE=InnoDB DEFAULT CHARSET=utf8;"
        table2 = "CREATE TABLE `people` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,\n`first_name` VARCHAR(255) NOT NULL DEFAULT '',\n`last_name` VARCHAR(255) NOT NULL DEFAULT '',\nPRIMARY KEY (`id`)) ENGINE=InnoDB DEFAULT CHARSET=utf8;"
        tables = [table1, table2]
        db = database_reader(tables)
        mygrate = mygration(db)
        ops = [str(op) for op in mygrate.operations]

        self.assertTrue(table1 in ops)
        self.assertTrue(table2 in ops)



import re

from .rule_children import rule_children
from .rule_delimited import rule_delimited
from .rule_literal import rule_literal
from .rule_regexp import rule_regexp
class parser(object):

    rule_types = {
        'children': rule_children,
        'delimited': rule_delimited,
        'literal': rule_literal,
        'regexp': rule_regexp
    }

    def __init__(self, rules=[]):

        self._values = {}

        # rules should be defined by the subclass
        if rules:
            self.rules = rules

        if not self.rules:
            raise NotImplementedError("Cannot extend parser without providing rules in %s" % (self.__class__))

        for (rule_index, rule) in enumerate(self.rules):

            # we always need a type
            if not 'type' in rule:
                raise ValueError('Missing type for rule %s in %s' % (rule, self.__class__))

        self.num_rules = len(self.rules)

    def __getitem__(self, key):

        return self._values[key]

    def __len__(self):

        return len(self._values)

    def has_key(self, key):

        return key in self._values

    def keys(self):

        return self._values.keys()

    def values(self):

        return self._values.values()

    def items(self):

        return self._values.items()

    def get_rule_parser(self, rule, next_rule):

        rule_type = rule['type']

        # keeping this simple for now
        if not rule_type in self.rule_types:
            raise ValueError('Unknown rule type %s for class %s' % (rule_type, self.__class__))

        return self.rule_types[rule_type](self, rule, next_rule)

    def parse(self, string=''):

        # first thing first, some initial string cleaning.  Clean spaces
        # from start and end and replace any multi-spaces with a single space.
        string = re.sub('\s+', ' ', string).strip()

        for (rule_index, rule) in enumerate(self.rules):

            # do we have a next rule?
            next_rule = self.rules[rule_index + 1] if rule_index < self.num_rules - 1 else False

            # now we can parse
            rule_parser = self.get_rule_parser(rule, next_rule)

            # does it match?  Check for a lack of match and deal with that first
            if not rule_parser.parse(string):

                # if this rule wasn't optional then we just don't match
                if not 'optional' in rule or not rule['optional']:
                    self.matched = False
                    return string

                # otherwise just keep going
                continue

            # we did match!  Yeah!
            self._values[rule_parser.name] = rule_parser.result
            string = rule_parser.leftovers

            # we are all done if we have nothing left
            if not string:
                break

        # if we got here then we got to the end, but we may not be done.  If we have more required
        # rules left that haven't been matched, then we don't match.

        # did we check every required rule?
        for check_index in range(rule_index + 1, self.num_rules):
            rule = self.rules[check_index]
            if not 'optional' in rule or not rule['optional']:
                self.matched = False
                return string

        # if we got here then we didn't match everything, but we fulfilled all of our
        # required rules.  As a result, we are done!
        self.process()
        self.matched = True
        return string

    def process(self):
        """ parser.process()

        Processes the results of the parsing process.  Only called if a match is found.  No input
        or output as it modifies the parser object in place, populating attributes as needed.
        """

        pass



import unittest

from mygrations.formats.mysql.file_reader.parsers.table_option import table_option
class test_table_options(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = table_option()
        returned = parser.parse("ENGINE=InnoDB")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals('', returned)

        # all we really have is the name and value
        self.assertEquals('ENGINE', parser.name)
        self.assertEquals('InnoDB', parser.value)

    def test_ignore_semicolon(self):

        # we don't handle the ending semi-colon
        parser = table_option()
        returned = parser.parse("COLLATE=utf8_general_ci;")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals(';', returned)

        # all we really have is the name and value
        self.assertEquals('COLLATE', parser.name)
        self.assertEquals('utf8_general_ci', parser.value)

    def test_ignore_spaces(self):

        # spaces are ignored as always
        parser = table_option()
        returned = parser.parse("COLLATE = utf8_general_ci")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals('', returned)

        # all we really have is the name and value
        self.assertEquals('COLLATE', parser.name)
        self.assertEquals('utf8_general_ci', parser.value)

    def test_ignore_spaces(self):

        # spaces are allowed in the name
        parser = table_option()
        returned = parser.parse("DEFAULT CHARSET=utf8")

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some data now
        self.assertEquals('', returned)

        # all we really have is the name and value
        self.assertEquals('DEFAULT CHARSET', parser.name)
        self.assertEquals('utf8', parser.value)



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_plain import type_plain
class test_type_plain(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = type_plain()
        returned = parser.parse("created date NOT NULL DEFAULT 'bob',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('created', parser.name)
        self.assertEquals('DATE', parser.column_type)
        self.assertFalse(parser.null)
        self.assertEquals('bob', parser.default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertEquals("`created` DATE NOT NULL DEFAULT 'bob'", str(parser))

    def test_optional_default(self):

        # parse typical insert values
        parser = type_plain()
        returned = parser.parse("created date")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('created', parser.name)
        self.assertEquals('DATE', parser.column_type)
        self.assertTrue(parser.null)
        self.assertEquals(None, parser.default)
        self.assertFalse(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertEquals("`created` DATE", str(parser))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = type_plain()
        returned = parser.parse("`created` date")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals('column', parser.definition_type)
        self.assertEquals('created', parser.name)
        self.assertEquals("`created` DATE", str(parser))

    def test_not_null_needs_default(self):

        # parse typical insert values
        parser = type_plain()
        returned = parser.parse("created date NOT NULL")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('is not null and has no default' in parser.warnings[0])



class change_constraint:
    """ Generates a partial SQL command to change a foreign key in a table """

    def __init__(self, constraint):
        self.constraint = constraint

    def apply_to_table(self, table):
        """ Changes the constraint in the table

        :param table: The table to change the constraint on
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.change_constraint(self.constraint)

    def __str__(self):
        return 'DROP FOREIGN KEY `%s`, ADD %s' % (self.constraint.name, str(self.constraint))



from .base import base
from mygrations.formats.mysql.file_reader.database import database as database_parser
from mygrations.formats.mysql.db_reader.database import database as database_reader
from mygrations.formats.mysql.mygrations.mygration import mygration
from mygrations.drivers.mysqldb.mysqldb import mysqldb
def execute(options):

    obj = import_files(options)
    obj.execute()
class import_files(base):
    def execute(self):

        files_database = database_parser(self.config['files_directory'])

        # any errors or warnings?
        if files_database.errors:
            print('Errors found in *.sql files')
            for error in files_database.errors:
                print(error)

            return False

        # use the credentials to load up a database connection
        live_database = database_reader(mysqldb(self.credentials))

        # we have to tell the live database to load records
        # for any tables we are tracking records for.
        # We use the "files" database as the "truth" because
        # just about any database can have records, but
        # that doesn't mean we want to track them.  We only
        # track them if the file has rows.
        for table in files_database.tables.values():
            if not table.tracking_rows:
                continue

            if not table.name in live_database.tables:
                continue

            live_database.read_rows(table)

        mygrate = mygration(files_database, live_database)
        if mygrate.errors_1215:
            print('1215 Errors encountered')
            for error in mygrate.errors_1215:
                print(error)

        else:
            for op in mygrate.operations:
                live_database.apply_to_source(op)



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_character import type_character
class test_type_character(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = type_character()
        returned = parser.parse("name varchar(255) not null default 'sup',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('name', parser.name)
        self.assertEquals('VARCHAR', parser.column_type)
        self.assertEquals('255', parser.length)
        self.assertFalse(parser.null)
        self.assertEquals('sup', parser.default)
        self.assertTrue(parser.has_comma)
        self.assertTrue(parser.character_set is None)
        self.assertTrue(parser.collate is None)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals("`name` VARCHAR(255) NOT NULL DEFAULT 'sup'", str(parser))

    def test_character_set(self):

        # parse typical insert values
        parser = type_character()
        returned = parser.parse("name varchar(255) not null default 'sup' character set 'blah' collate 'boo',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('name', parser.name)
        self.assertEquals('VARCHAR', parser.column_type)
        self.assertEquals('255', parser.length)
        self.assertFalse(parser.null)
        self.assertEquals('sup', parser.default)
        self.assertTrue(parser.has_comma)
        self.assertEquals('BLAH', parser.character_set)
        self.assertEquals('BOO', parser.collate)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals("`name` VARCHAR(255) NOT NULL DEFAULT 'sup' CHARACTER SET 'BLAH' COLLATE 'BOO'", str(parser))

    def test_optional_default(self):

        # parse typical insert values
        parser = type_character()
        returned = parser.parse("name varchar(255)")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('name', parser.name)
        self.assertEquals('VARCHAR', parser.column_type)
        self.assertEquals('255', parser.length)
        self.assertTrue(parser.null)
        self.assertEquals(None, parser.default)
        self.assertFalse(parser.has_comma)
        self.assertTrue(parser.character_set is None)
        self.assertTrue(parser.collate is None)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals("`name` VARCHAR(255)", str(parser))

    def test_empty_default(self):

        # parse typical insert values
        parser = type_character()
        returned = parser.parse("name varchar(255) default ''")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('', parser.default)
        self.assertEquals("`name` VARCHAR(255) DEFAULT ''", str(parser))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = type_character()
        returned = parser.parse("`name` varchar(255)")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('column', parser.definition_type)
        self.assertEquals('name', parser.name)
        self.assertEquals("`name` VARCHAR(255)", str(parser))

    def test_not_null_needs_default(self):

        # parse typical insert values
        parser = type_character()
        returned = parser.parse("name varchar(255) NOT NULL")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertTrue('is not null and has no default' in parser.warnings[0])

    # only char and varchar can have collate
    def test_disallowed_collate(self):

        # parse typical insert values
        parser = type_character()
        returned = parser.parse("name int(10) collate 'blah'")
        self.assertTrue('is not allowed to have a collation or character set' in parser.errors[0])

    def test_disallowed_character_set(self):

        # parse typical insert values
        parser = type_character()
        returned = parser.parse("name int(10) character set 'blah'")
        self.assertTrue('is not allowed to have a collation or character set' in parser.errors[0])

    # some types are not allowed to have a length
    def test_invalid_column_type(self):

        for coltype in [
            'date', 'year', 'tinyblob', 'blob', 'mediumblob', 'longblob', 'tinytext', 'text', 'mediumtext', 'longtext',
            'json'
        ]:

            parser = type_character()
            returned = parser.parse('name %s(10)' % coltype)

            self.assertTrue(parser.matched)
            self.assertEquals('', returned)
            self.assertTrue('not allowed to have a length' in parser.errors[0])



class remove_constraint:
    """ Generates a partial SQL command to drop a foreign key from a table """

    def __init__(self, constraint):
        self.constraint = constraint

    def apply_to_table(self, table):
        """ Removes the constraint from the table

        :param table: The table to remove the constraint from
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.remove_constraint(self.constraint)

    def __str__(self):
        return 'DROP FOREIGN KEY `%s`' % (self.constraint.name)



import unittest

from mygrations.formats.mysql.file_reader.database import database as database_reader
from mygrations.formats.mysql.file_reader.create_parser import create_parser
class test_database_unfulfilled_fks(unittest.TestCase):
    def test_can_fulfill(self):
        """ If we have all necessary columns in all necessary tables, we can fulfill the FKs """

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """, """
            CREATE TABLE `types` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `name` VARCHAR(255),
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `type_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            KEY `log_changes_type_id` (`type_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            CONSTRAINT `log_changes_type_id_fk` FOREIGN KEY (`type_id`) REFERENCES `types` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
            );
        """
        )

        self.assertEquals(0, len(db.unfulfilled_fks(new_table)))

    def test_can_not_fulfill(self):
        """ If we are missing even just one we can't fulfill and should get back the missing constraint """

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `type_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            KEY `log_changes_type_id` (`type_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            CONSTRAINT `log_changes_type_id_fk` FOREIGN KEY (`type_id`) REFERENCES `types` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_type_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_type_id_fk` FOREIGN KEY (`type_id`) REFERENCES `types` (`id`) ON DELETE CASCADE ON UPDATE CASCADE',
            str(errors_1215['log_changes_type_id_fk']['foreign_key'])
        )
        self.assertEquals(
            'MySQL 1215 error for foreign key `log_changes_type_id_fk`: `log_changes`.`type_id` references `types`.`id`, but table `types` does not exist',
            errors_1215['log_changes_type_id_fk']['error']
        )

    def test_missing_column(self):
        """ If we are missing even just one we can't fulfill and should get back the missing constraint """

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `type_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            KEY `log_changes_type_id` (`type_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            CONSTRAINT `log_changes_type_id_fk` FOREIGN KEY (`type_id`) REFERENCES `logs` (`bob`) ON DELETE CASCADE ON UPDATE CASCADE
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_type_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_type_id_fk` FOREIGN KEY (`type_id`) REFERENCES `logs` (`bob`) ON DELETE CASCADE ON UPDATE CASCADE',
            str(errors_1215['log_changes_type_id_fk']['foreign_key'])
        )
        self.assertEquals(
            'MySQL 1215 error for foreign key `log_changes_type_id_fk`: `log_changes`.`type_id` references `logs`.`bob`, but column `logs`.`bob` does not exist',
            errors_1215['log_changes_type_id_fk']['error']
        )

    def test_1215_wrong_type(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` BIGINT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_log_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE',
            str(errors_1215['log_changes_log_id_fk']['foreign_key'])
        )
        self.assertEquals(
            "MySQL 1215 error for foreign key `log_changes_log_id_fk`: column type mismatch. `log_changes`.`log_id` is 'BIGINT' but `logs`.`id` is 'INT'",
            errors_1215['log_changes_log_id_fk']['error']
        )

    def test_1215_wrong_length(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(11) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_log_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE',
            str(errors_1215['log_changes_log_id_fk']['foreign_key'])
        )
        self.assertEquals(
            "MySQL 1215 error for foreign key `log_changes_log_id_fk`: length mismatch. `log_changes`.`log_id` is '11' but `logs`.`id` is '10'",
            errors_1215['log_changes_log_id_fk']['error']
        )

    def test_1215_wrong_unsigned(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_log_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE',
            str(errors_1215['log_changes_log_id_fk']['foreign_key'])
        )
        self.assertEquals(
            "MySQL 1215 error for foreign key `log_changes_log_id_fk`: unsigned mistmatch. `logs`.`id` is unsigned but `log_changes`.`log_id` is not",
            errors_1215['log_changes_log_id_fk']['error']
        )

    def test_1215_wrong_unsigned_reversed(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_log_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE',
            str(errors_1215['log_changes_log_id_fk']['foreign_key'])
        )
        self.assertEquals(
            "MySQL 1215 error for foreign key `log_changes_log_id_fk`: unsigned mistmatch. `log_changes`.`log_id` is unsigned but `logs`.`id` is not",
            errors_1215['log_changes_log_id_fk']['error']
        )

    def test_1215_on_delete_set_null_not_nullable(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) UNSIGNED NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE SET NULL ON UPDATE CASCADE,
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_log_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE SET NULL ON UPDATE CASCADE',
            str(errors_1215['log_changes_log_id_fk']['foreign_key'])
        )
        self.assertEquals(
            "MySQL 1215 error for foreign key `log_changes_log_id_fk`: invalid SET NULL. `log_changes`.`log_id` is not allowed to be null but the foreign key attempts to set the value to null ON DELETE",
            errors_1215['log_changes_log_id_fk']['error']
        )

    def test_1215_on_update_set_null_not_nullable(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) UNSIGNED NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE SET NULL,
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_log_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE SET NULL',
            str(errors_1215['log_changes_log_id_fk']['foreign_key'])
        )
        self.assertEquals(
            "MySQL 1215 error for foreign key `log_changes_log_id_fk`: invalid SET NULL. `log_changes`.`log_id` is not allowed to be null but the foreign key attempts to set the value to null ON UPDATE",
            errors_1215['log_changes_log_id_fk']['error']
        )

    def test_1215_on_update_on_delete_set_null_not_nullable(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) UNSIGNED NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE SET NULL ON UPDATE SET NULL,
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_log_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE SET NULL ON UPDATE SET NULL',
            str(errors_1215['log_changes_log_id_fk']['foreign_key'])
        )
        self.assertEquals(
            "MySQL 1215 error for foreign key `log_changes_log_id_fk`: invalid SET NULL. `log_changes`.`log_id` is not allowed to be null but the foreign key attempts to set the value to null ON DELETE and ON UPDATE",
            errors_1215['log_changes_log_id_fk']['error']
        )

    def test_missing_index(self):
        """ If we have all necessary columns in all necessary tables, we can fulfill the FKs """

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """, """
            CREATE TABLE `types` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `name` VARCHAR(255),
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        db = database_reader(strings)

        new_table = create_parser()
        new_table.parse(
            """CREATE TABLE `log_changes` (
            `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
            `log_id` INT(10) UNSIGNED NOT NULL,
            `type_id` INT(10) UNSIGNED NOT NULL,
            `change` VARCHAR(255),
            PRIMARY KEY (id),
            KEY `log_changes_log_id` (`log_id`),
            KEY `log_changes_type_id` (`type_id`),
            CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
            CONSTRAINT `log_changes_type_id_fk` FOREIGN KEY (`type_id`) REFERENCES `types` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
            );
        """
        )

        errors_1215 = db.unfulfilled_fks(new_table)
        self.assertEquals(1, len(errors_1215))
        self.assertTrue('log_changes_log_id_fk' in errors_1215)
        self.assertEquals(
            'CONSTRAINT `log_changes_log_id_fk` FOREIGN KEY (`log_id`) REFERENCES `logs` (`id`) ON DELETE CASCADE ON UPDATE CASCADE',
            str(errors_1215['log_changes_log_id_fk']['foreign_key'])
        )
        self.assertEquals(
            "MySQL 1215 error for foreign key `log_changes_log_id_fk`: missing index. `log_changes`.`log_id` references `logs`.`id` but `logs`.`id` does not have an index and therefore cannot be used in a foreign key constraint",
            errors_1215['log_changes_log_id_fk']['error']
        )



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
from mygrations.formats.mysql.definitions.index import index
class test_table_operations_key(unittest.TestCase):
    def _get_default_table(self):
        table = create_parser()
        table.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            `subject` varchar(255) NOT NULL DEFAULT '',
            PRIMARY KEY (id),
            KEY `tasks_account_id` (`account_id`),
            CONSTRAINT `tasks_account_id_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """
        )

        return table

    def test_add_key(self):
        table = self._get_default_table()
        self.assertTrue(table.column_is_indexed('account_id'))

        table.add_key(index('task_tasks', ['task', 'subject']))
        self.assertTrue('task_tasks' in table.indexes)
        self.assertTrue(table.column_is_indexed('task'))

    def test_cannot_add_duplicate_key(self):
        table = self._get_default_table()
        with self.assertRaises(ValueError):
            table.add_key(index('tasks_account_id', ['task', 'subject']))

    def test_remove_key(self):
        table = self._get_default_table()
        self.assertTrue(table.column_is_indexed('account_id'))

        table.remove_key('tasks_account_id')
        self.assertFalse('tasks_account_id' in table.indexes)
        self.assertFalse(table.column_is_indexed('account_id'))

    def test_cannot_remove_missing_key(self):
        table = self._get_default_table()
        with self.assertRaises(ValueError):
            table.remove_key('asdf')

    def test_change_key(self):
        table = self._get_default_table()
        self.assertTrue(table.column_is_indexed('account_id'))

        table.change_key(index('tasks_account_id', ['account_id', 'task']))
        self.assertEquals('KEY `tasks_account_id` (`account_id`,`task`)', str(table.indexes['tasks_account_id']))
        self.assertTrue(table.column_is_indexed('account_id'))

    def test_cannot_change_missing_key(self):
        table = self._get_default_table()
        with self.assertRaises(ValueError):
            table.change_key(index('asdf', ['account_id', 'task']))



__ver__ = '1.0'

import io
import os
import re

import json
from json import JSONDecodeError
class DotEnvSyntaxError(Exception):
    pass
class dotenv(dict):

    raw_contents = ''
    comment = ''
    comment_length = 0

    def __init__(self, filename='', comment='#'):

        # we are going to allow a few different values for filename.
        # we will allow an actual string to parse, a string as a filename,
        # a file object, or an IO stream.  get_contents will handle the
        # details
        if not os.path.isfile(filename):
            raise ValueError("Specified file '%s' does not exist" % filename)
        if filename:
            self.raw_contents = self.get_contents(filename)

            # and parse
            parsed = self.parse(self.raw_contents, comment)

        else:
            self.raw_contents = ''
            parsed = {}

        super().__init__(parsed)

    def get_contents(self, filename):

        # file pointer
        if isinstance(filename, io.IOBase):
            contents = filename.read()

        # and an actual string
        elif isinstance(filename, str):

            # which could be a filename
            if os.path.isfile(filename):
                fp = open(filename, 'r')
                contents = fp.read()
                fp.close()
            else:
                contents = filename

        else:
            raise ValueError(
                "Unknown type for filename: must be an ascii string, a filename, file pointer, or StringIO"
            )

        # we didn't end up with bytes, did we?
        try:
            contents = contents.decode()
        except AttributeError:
            pass

        return contents

    def parse(self, string, comment='#'):

        # this isn't json, is it?
        parsed = False
        try:
            parsed = json.loads(string)
        except JSONDecodeError:
            pass

        # because if it is, this is very easy
        if parsed:
            return parsed

        # This will hold our results
        configs = {}

        # okay then!  Let's get started.  Divide up by newlines and work from there
        for line in string.split("\n"):

            # we're not actually going to do much here
            [key, value] = self.parse_line(line, comment)

            # we will get a non-value for key and value if this wasn't
            # a valid data line
            if not key or not value:
                continue

            # store it!
            configs[key] = value

        # that's it
        return configs

    def parse_line(self, line, comment):

        # clean it
        clean = line.strip()

        # see if we start with a comment character
        if clean[:len(comment)] == comment:
            return [False, False]

        # also ignore empty lines
        if not clean:
            return [False, False]

        # we want to do actual parsing, instead of doing
        # it the easy way and just splitting on an equal sign.
        # this is mainly intended to give some flexibility to support
        # a few different potentially common formats.  For instance,
        # an equal sign is a common delimiter between key and value,
        # but a colon (for a JSON-like syntax) may also be used.
        match = re.match('^([^=:]+)[=:](.*)', clean)

        # if that didn't match anything then we don't have a config
        if not match:
            raise DotEnvSyntaxError("Syntax error: no data found in line: %s" % line)

        # otherwise we will continue on our way.  Clean everything up.
        key = match.group(1).strip()
        value = match.group(2).strip()

        # non-value?
        if not value:
            return [key, '']

        # our key should be good to go.  Our value needs some checking.
        # it may be surrounded by quotes (single or double).  It may
        # have a comment character at the end, or it may have a comment
        # character inside a quoted string which is not actually a comment.
        # therefore, we need some fancy parsing.

        # the main starting question is whether or not our value starts
        # with a quote.  If not this is easy.
        if value[0] != "'" and value[0] != '"':

            # just check for a comment
            # and remove it if found
            if value.count(comment):
                value = value[:value.index(comment)].strip()

            # if we have a quote character elsewhere in the value though,
            # we have a problem
            if value.count('"') or value.count("'"):
                raise DotEnvSyntaxError("Syntax error: unclosed quote in line %s" % line)

            return [key, value]

        # if we got here then we have a quoted string.  We want to validate the syntax,
        # remove the quotes, check for escaped quotes, and check for comment
        quote = value[0]
        value = value[1:]

        # syntax error?
        if not value.count(quote):
            raise DotEnvSyntaxError("Syntax error: unclosed quote in line %s" % line)

        # do we have an empty string?
        if value[0] == quote:
            return [key, '']

        # make sure we don't have an escaped quote in our quoted string
        index = 0
        while value.count(quote, index):
            index = value.index(quote, index + 1)
            if value[index - 1] == '\\':
                continue

            # if we get here then we have found the ending quote!

            # check for a comment
            if value.count(comment, index):
                value = value[:value.index(comment, index)].strip()

            # our string should end at the closing quote, although safely ignore a closing semi-colon
            if len(value) > index + 1 and value[index + 1:].strip() != ';':
                raise DotEnvSyntaxError("Syntax error: data found outside of the quote in line %s" % line)

            # don't forget to unescape any quotes and return
            return [key, value[:index].replace('\\%s' % quote, quote)]

        # if we got here then we didn't find anything
        raise DotEnvSyntaxError("Syntax error: unclosed quote in line %s" % line)



class add_constraint:
    """ Generates a partial SQL command to add a FK to a table """

    def __init__(self, constraint):
        self.constraint = constraint

    def apply_to_table(self, table):
        """ Adds the constraint to the table

        :param table: The table to add the constraint to
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.add_constraint(self.constraint)

    def __str__(self):
        return 'ADD %s' % str(self.constraint)



import MySQLdb
import MySQLdb.cursors

from collections import OrderedDict
class mysqldb(object):
    """ High level driver for a MySQLdb connection """
    execute_cursor = None

    def __init__(self, credentials):
        """ Initialize the MySQLdb connection

        Accepts either a dictionary with mysqldb connnection credentials
        or any MySQL database connection which implements the Python DB API spec v2.0

        :param credentials: The database credentials to connect with or a database connection
        :type credentials: dict|mygrations.helpers.db_credentials
        """
        if issubclass(type(credentials), dict):
            self.conn = MySQLdb.connect(**credentials)
        else:
            self.conn = credentials

    def tables(self):
        """ Returns the tables in the connected database

        :returns: An ordered dict with table definitions by table name
        :rtype: OrderedDict
        """
        cursor = self.conn.cursor()

        # not returning an iterator: just fetch everything.
        # I'm guessing this will be fine for any realistic database
        # size, and avoids issues of having multiple open cursors
        # at the same time.
        cursor.execute('SHOW TABLES')
        table_names = []
        for (table_name, ) in cursor:
            table_names.append(table_name)

        definitions = OrderedDict()
        for table_name in table_names:
            cursor.execute('SHOW CREATE TABLE %s' % table_name)
            if not cursor.rowcount:
                raise ValueError("Failed to execute SHOW CREATE TABLE command on table %s" % table_name)

            (tbl_name, create_table) = cursor.fetchone()
            definitions[table_name] = create_table

        cursor.close()

        return definitions

    def rows(self, table_name):
        """ Returns the rows in the table as a tuple of dicts

        :returns: The rows in the table
        :rtype: (dict)
        """

        # still sticking to full caching client-side without
        # support for memory-preserving iterators.  Given our
        # use case, I think this will be fine.  Can always change later
        cursor = self.conn.cursor(MySQLdb.cursors.DictCursor)
        cursor.execute('SELECT * FROM %s' % table_name)
        rows = cursor.fetchall()
        cursor.close()

        return rows

    def execute(self, queries):
        if isinstance(queries, str):
            queries = [queries]

        cursor = self.conn.cursor(MySQLdb.cursors.DictCursor)
        for query in queries:
            cursor.execute(query)
        self.conn.commit()
        cursor.close()



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
from mygrations.formats.mysql.definitions.column import column
class test_table_operations_column(unittest.TestCase):
    def _get_default_table(self):
        table = create_parser()
        table.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            `subject` varchar(255) NOT NULL DEFAULT '',
            PRIMARY KEY (id),
            KEY `tasks_account_id` (`account_id`),
            CONSTRAINT `tasks_account_id_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """
        )

        return table

    def _get_new_column(self):
        return column(**{"name": "description", "column_type": "TEXT", "length": ""})

    def test_add_column_before(self):
        table = self._get_default_table()
        table.add_column(self._get_new_column(), 'task')

        self.assertTrue('description' in table.columns)
        self.assertEquals('task', table.column_before('description'))

    def test_add_column_first(self):
        table = self._get_default_table()
        table.add_column(self._get_new_column(), True)

        self.assertTrue('description' in table.columns)
        self.assertTrue(table.column_before('description'))

    def test_add_column_last(self):
        table = self._get_default_table()
        table.add_column(self._get_new_column(), False)

        self.assertTrue('description' in table.columns)
        self.assertEquals('subject', table.column_before('description'))

    def test_add_column_after_non_existent(self):
        table = self._get_default_table()
        with self.assertRaises(ValueError):
            table.add_column(self._get_new_column(), 'asdfef')

    def test_cannot_add_duplicate_column(self):
        table = self._get_default_table()
        table.add_column(self._get_new_column())
        with self.assertRaises(ValueError):
            table.add_column(self._get_new_column())

    def test_remove_column(self):
        table = self._get_default_table()
        table.remove_column('task')
        self.assertFalse('task' in table.columns)

    def test_change_column(self):
        table = self._get_default_table()
        new_subject = column(**{"name": "subject", "column_type": "TEXT", "length": ""})
        table.change_column(new_subject)

        self.assertEquals('`subject` TEXT', str(table.columns['subject']))

    def test_change_column_no_rename(self):
        table = self._get_default_table()
        with self.assertRaises(ValueError):
            table.change_column(self._get_new_column())



class remove_column:
    """ Generates a partial SQL command to remove a column from a table """

    def __init__(self, column):
        self.column = column

    def apply_to_table(self, table):
        """ Removes the column from the table

        :param table: The table to remove the column from
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.remove_column(self.column)

    def __str__(self):
        return 'DROP %s' % self.column.name



class add_key:
    """ Generates a partial SQL command to add a key to a table """

    def __init__(self, key):
        self.key = key

    def apply_to_table(self, table):
        """ Adds the key to the table

        :param table: The table to add the key to
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.add_key(self.key)

    def __str__(self):
        return 'ADD %s' % str(self.key)



import unittest
from mygrations.formats.mysql.mygrations.operations.row_update import row_update
from collections import OrderedDict
class test_row_update(unittest.TestCase):
    def test_simple(self):
        op = row_update('a_table', OrderedDict([('id', 5), ('name', 'bob'), ('age', 2)]))

        self.assertEquals("UPDATE `a_table` SET `name`='bob', `age`='2' WHERE id=5;", str(op))
        self.assertEquals('a_table', op.table_name)

    def test_id_required(self):
        with self.assertRaises(KeyError):
            op = row_update('a_table', OrderedDict([('name', 'bob'), ('age', 2)]))



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_enum import type_enum
class test_type_enum(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = type_enum()
        returned = parser.parse("options enum('bob','joe','ray') not null default 'bob',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('options', parser.name)
        self.assertEquals('ENUM', parser.column_type)
        self.assertEquals(['bob', 'joe', 'ray'], parser.values)
        self.assertFalse(parser.null)
        self.assertEquals('bob', parser.default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertEquals("`options` ENUM('bob','joe','ray') NOT NULL DEFAULT 'bob'", str(parser))

    def test_optional_default(self):

        # parse typical insert values
        parser = type_enum()
        returned = parser.parse("options enum('bob','joe','ray')")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('options', parser.name)
        self.assertEquals('ENUM', parser.column_type)
        self.assertEquals(['bob', 'joe', 'ray'], parser.values)
        self.assertTrue(parser.null)
        self.assertEquals(None, parser.default)
        self.assertFalse(parser.has_comma)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(0, len(parser.warnings))
        self.assertEquals("`options` ENUM('bob','joe','ray')", str(parser))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = type_enum()
        returned = parser.parse("`options` enum('bob','joe','ray')")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals('options', parser.name)

    def test_warning_for_no_quote_default(self):

        # parse typical insert values
        parser = type_enum()
        returned = parser.parse("options enum('bob','joe','ray') default bob")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('should have quotes' in parser.warnings[0])
        self.assertEquals("`options` ENUM('bob','joe','ray') DEFAULT 'bob'", str(parser))

    def test_error_for_invalid_default(self):

        # parse typical insert values
        parser = type_enum()
        returned = parser.parse("options enum('bob','joe','ray') default asdf")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('this is not an allowed value' in parser.errors[0])

    def test_not_null_needs_default(self):

        # parse typical insert values
        parser = type_enum()
        returned = parser.parse("options enum('bob','joe','ray') not null")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertTrue('is not null and has no default' in parser.warnings[0])

    # only some can match this type
    def test_allowed_types(self):

        # parse typical insert values
        for coltype in ['set', 'enum']:
            parser = type_enum()
            returned = parser.parse("name %s('bob','joe','ray')" % coltype)
            self.assertEquals(
                0, len(parser.errors), 'Type %s should be allowed to use set/enum values but is not' % coltype
            )

    # everything else is not allowed to have a length
    def test_invalid_column_type(self):

        for coltype in [
            'date', 'year', 'tinyblob', 'blob', 'mediumblob', 'longblob', 'tinytext', 'text', 'mediumtext', 'longtext',
            'json', 'char', 'varchar', 'biejfiejrei'
        ]:

            parser = type_enum()
            returned = parser.parse("name %s('bob','joe','ray')" % coltype)

            self.assertTrue(parser.matched)
            self.assertEquals('', returned)
            self.assertTrue(
                'not allowed to have a list of values' in parser.errors[0],
                'Type %s should not be allowed to use set/enum but is' % coltype
            )



class rows(object):

    _table = ''
    _raw_rows = ''
    _columns = None
    _errors = None
    _warnings = None

    @property
    def table(self):
        """ Public getter.  Returns the name of the table that records are being inserted for

        :returns: The name of the table
        :rtype: list
        """

        return self._table

    @property
    def raw_rows(self):
        """ Public getter.  Returns a list of insert values from the VALUES part of the query

        :returns: A list of values for each row
        :rtype: list[list]
        """

        return self._raw_rows

    @property
    def columns(self):
        """ Public getter.  Returns the list of columns for the rows

        :returns: A list of columns
        :rtype: list
        """
        return self._columns

    @property
    def num_explicit_columns(self):
        """ Public getter.  Returns the number of columns specified in the insert query

        Can be zero if none are specified, which happens for queries like:
        INSERT INTO table (val1, val2, val3...);

        :returns: The number of columns which have been explicitly defined
        :rtype: integer
        """
        return self._num_explicit_columns

    @property
    def errors(self):
        """ Public getter.  Returns a list of parsing errors

        :returns: A list of parsing errors
        :rtype: list
        """
        return [] if self._errors is None else self._errors

    @property
    def warnings(self):
        """ Public getter.  Returns a list of parsing/table warnings

        :returns: A list of parsing/table warnings
        :rtype: list
        """
        return [] if self._warnings is None else self._warnings



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
class test_constraint_foreign(unittest.TestCase):
    def test_complicated_table_parses(self):

        # parse a typical foreign key constraint
        parser = create_parser()
        returned = parser.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `membership_id` int(10) DEFAULT NULL,
            `status_id` int(10) DEFAULT NULL,
            `priority_id` int(10) unsigned DEFAULT NULL,
            `task_type_id` int(10) unsigned DEFAULT NULL,
            `task_team_id` int(10) unsigned DEFAULT NULL,
            `repeating_tasks_id` int(10) unsigned DEFAULT NULL,
            `subject` varchar(255) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            `due_date` int(10) DEFAULT NULL,
            `original_due_date` int(10) DEFAULT NULL,
            `assigned_to_id` int(10) unsigned DEFAULT NULL,
            `delegated_to_id` int(10) DEFAULT NULL,
            `trust` tinyint(1) NOT NULL DEFAULT '0',
            `description` text,
            `multiple_task_id` int(10) DEFAULT NULL,
            `completed_dt` int(10) DEFAULT NULL,
            `duration` int(10) NOT NULL DEFAULT '0',
            `number_comments` int(10) NOT NULL DEFAULT '0',
            `number_uploads` int(10) NOT NULL DEFAULT '0',
            `created` int(10) DEFAULT NULL,
            `updated` int(10) DEFAULT NULL,
            PRIMARY KEY (`id`),
            KEY `task_status_id` (`status_id`),
            KEY `task_priority_id` (`priority_id`),
            KEY `tasks_membership_id` (`membership_id`),
            KEY `task_type_id` (`task_type_id`),
            KEY `task_assigned_to_id` (`assigned_to_id`),
            CONSTRAINT `tasks_assigned_to_id_ref_memberships_user_id` FOREIGN KEY (`assigned_to_id`) REFERENCES `memberships` (`id`) ON DELETE SET NULL,
            CONSTRAINT `tasks_priority_id_ref_task_priorities_id` FOREIGN KEY (`priority_id`) REFERENCES `task_priorities` (`id`) ON UPDATE CASCADE,
            CONSTRAINT `tasks_type_id_ref_task_types_id` FOREIGN KEY (`task_type_id`) REFERENCES `task_types` (`id`) ON UPDATE CASCADE
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        )

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have matched everything
        self.assertEquals('tasks', parser.name)
        self.assertEquals('', returned)
        self.assertEquals(23, len(parser.columns))
        self.assertEquals(6, len(parser.indexes))
        self.assertEquals(3, len(parser.constraints))
        self.assertEquals(2, len(parser.options))
        self.assertTrue(parser.semicolon)
        self.assertEquals(0, len(parser.errors))
        self.assertEquals(['id'], parser.primary.columns)

    def test_keeps_errors(self):

        # parse a typical foreign key constraint
        parser = create_parser()
        returned = parser.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `subject` varchar DEFAULT NULL,
            `task` text DEFAULT NULL
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        )

        # we should have matched
        self.assertTrue(parser.matched)

        # and we should have some errors
        self.assertTrue('must have a length for column subject in table tasks' in parser.errors[0])
        self.assertTrue('not allowed to have a default value for column task in table tasks' in parser.errors[1])



from .rule_base import rule_base
import re
class rule_regexp(rule_base):
    """ rule = rule_regexp( parser, rule, next_rule )

    rule_literal constructor.  Pass in the parser object that the rule is for, the rule
    that is being parsed, and the next rule in the parsing chain.

    This class attempts to match a regular expression (case-insensitive).

    keys for the rules dictionary:

    =========  ========
    key        contents
    =========  ========
    value      The regular expression to look for
    name       The name of the rule
    optional   (optional) If true, denotes that this rule is optional
    =========  ========

    :param parser: The parser that this rule belongs to
    :param rule: A dictionary representing the rule configuration (see each rule class for details)
    :param next_rule: A dictionary representing the next rule
    :type parser: parser
    :type rule: dict
    :type next_rule: dict
    :returns: rule object
    """

    regexp = ''

    def __init__(self, parser, rule, next_rule):

        super().__init__(parser, rule, next_rule)

        self.regexp = self.rule['value']

    def parse(self, string):

        # apply the regular expression!
        result = re.match(self.regexp, string, re.IGNORECASE)

        # if it didn't match then nothing channged
        if not result:
            self.leftovers = string
            self.result = ''
            return False

        # otherwise we have a match and can return as such
        self.result = result.group(0)
        cleaned = string[len(self.result):].strip()

        # if there was a group in the regular expression then just keep that part
        if result.groups():
            self.result = result.group(1)

        self.leftovers = cleaned
        return True



from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.constraint import constraint
class constraint_foreign(parser, constraint):

    has_comma = False

    # CONSTRAINT `accounts_status_id_ref_account_statuses_id` FOREIGN KEY (`status_id`) REFERENCES `account_statuses` (`id`) ON UPDATE CASCADE
    rules = [{
        'type': 'literal',
        'value': 'CONSTRAINT'
    }, {
        'type': 'regexp',
        'name': 'name',
        'value': '[^\(\s\)]+'
    }, {
        'type': 'literal',
        'value': 'FOREIGN KEY ('
    }, {
        'type': 'regexp',
        'name': 'column',
        'value': '[^\(\s\)]+'
    }, {
        'type': 'literal',
        'value': ') REFERENCES'
    }, {
        'type': 'regexp',
        'name': 'foreign_table',
        'value': '[^\(]+'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'regexp',
        'name': 'foreign_column',
        'value': '[^\)]+'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': 'ON DELETE CASCADE',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON DELETE NO ACTION',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON DELETE RESTRICT',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON DELETE SET DEFAULT',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON DELETE SET NULL',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON UPDATE CASCADE',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON UPDATE NO ACTION',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON UPDATE RESTRICT',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON UPDATE SET DEFAULT',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'ON UPDATE SET NULL',
        'optional': True
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):

        self._errors = []
        self._warnings = []
        self._name = self._values['name'].strip().strip('`')
        self._column = self._values['column'].strip().strip('`')
        self._foreign_table = self._values['foreign_table'].strip().strip('`')
        self._foreign_column = self._values['foreign_column'].strip().strip('`')

        # figure out what our rules are
        self._on_delete = self.find_action('DELETE')
        self._on_update = self.find_action('UPDATE')

        self.has_comma = True if 'ending_comma' in self._values else False

        if len(self._name) > 64:
            self._errors.append('Key name %s is too long' % (self._name))

    def find_action(self, update_type):

        # watch for more than one action for this type
        found = ''
        for action in ['CASCADE', 'NO ACTION', 'RESTRICT', 'SET DEFAULT', 'SET NULL']:
            if not ('ON %s %s' % (update_type, action)) in self._values:
                continue

            if found:
                self._errors.append('Found contradictory rules in foreign constraint for ON %s' % update_type)
            else:
                found = action

        # restrict is the default for MySQL if not specified
        if not found:
            return 'RESTRICT'

        return found.upper()



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
class test_table_difference(unittest.TestCase):
    def test_simple_create(self):
        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id)
            );
        """
        )

        self.assertEquals(
            str(a).replace("\n", ' '),
            "CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT, `account_id` INT(10), `task` VARCHAR(255), PRIMARY KEY (`id`));"
        )

    def test_with_constraints(self):
        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            PRIMARY KEY (id),
            CONSTRAINT `tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE
            );
        """
        )

        self.assertEquals(
            str(a).replace("\n", ' '),
            "CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT, `account_id` INT(10), PRIMARY KEY (`id`), CONSTRAINT `tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE RESTRICT);"
        )

    def test_with_options(self):
        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL,
            PRIMARY KEY (id)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        )

        self.assertEquals(
            str(a).replace("\n", ' '),
            "CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT, `account_id` INT(10), `task` VARCHAR(255), PRIMARY KEY (`id`)) ENGINE=InnoDB DEFAULT CHARSET=utf8;"
        )



import copy

from mygrations.formats.mysql.definitions.database import database
from mygrations.formats.mysql.mygrations.operations.alter_table import alter_table
from mygrations.formats.mysql.mygrations.operations.add_constraint import add_constraint
from mygrations.formats.mysql.mygrations.operations.remove_table import remove_table
from mygrations.formats.mysql.mygrations.operations.disable_checks import disable_checks
from mygrations.formats.mysql.mygrations.operations.enable_checks import enable_checks
class mygration:
    """ Creates a migration plan to update a database to a given spec.

    If only one database is passed in, it treats it as a structure to migrate
    to starting from a blank slate, which primariliy means just figuring out
    what order to add tables in to account for foreign key constraints.

    If two tables are present, then it will treat the second as the current
    database constraint, and will figure out steps to execute in action
    to update that second database to match the structure of the first.

    The general steps are:

    1. Check for foreign key errors: if the two database we are trying to migrate to doesn't support its own foreign key constraints, there is a show stopper
    2. Loop through tables to be added, "creating" them as their foreign key constraints are filled, and ignoring any that have interlocking dependencies
    3. Remove any FK constraints that are no longer used (done early to prevent trouble in the next step)
    4. Modify tables as needed.  If modifying tables involves adding foreign key constraints that are not fulfilled, ignore those and add them later
    5. See if we can add any remaining tables now that some table modifications have been applied
    6. If there are still any outstanding tables to add, remove any foreign key constraints that are not fulfilled and add the tables without them
    7. Apply all foreign key constraints that were previously ignored in steps 4 & 6
    8. Remove any tables that need to be removed
    """

    def __init__(self, db_to, db_from=None, disable_checks=True):
        """ Create a migration plan

        :param db_to: The target database structure to migrate to
        :param db_from: The current database structure to migrate from
        :param disable_checks: Whether or not to add an operations to disable/re-enable FK checks
        :type db_to: mygrations.formats.mysql.definitions.database
        :type db_from: mygrations.formats.mysql.definitions.database
        :type disable_checks: True
        """

        self.db_to = db_to
        self.db_from = db_from
        self._disable_fk_checks = disable_checks

        # first things first: stop if we have any FK errors in the database
        # we are migrating to
        self._errors_1215 = []
        self._errors_1215 = self.db_to.errors_1215

        if not self._errors_1215:
            self._operations = self._process()
        else:
            self._operations = None

    @property
    def operations(self):
        """ Public getter.  Returns list of operations to bring db_from to db_to

        If db_from doesn't exist then it will be a list of operations to
        create db_to.

        If there are 1215 errors that will prevent the migration, then
        operations will return None

        If the two databases are exactly the same then operations will
        be an empty list

        :returns: A list of table operations
        :rtype: None|[mygrations.formats.mysql.mygrations.operations.operation]
        """
        return self._operations

    @property
    def errors_1215(self):
        """ Public getter.  Returns list of 1215 errors (as strings)

        :returns: A list of 1215 error messages
        :rtype: [string]
        """
        return self._errors_1215

    def __len__(self):
        return len(self._operations)

    def __bool__(self):
        return True if len(self._operations) else False

    def __str__(self):
        return "\n".join([str(x) for x in self._operations])

    def __iter__(self):
        return self._operations.__iter__()

    def _differences(self, from_dict, to_dict):
        """
        Calculates the difference between two OrderedDicts.

        https://codereview.stackexchange.com/a/176303/140581

        Duplication!!!! (formats.mysql.create_parser).  Sue me.

        :param a: OrderedDict
        :param b: OrderedDict
        :return: (added, removed, overlap)
        """

        return ([key for key in to_dict if key not in from_dict], [key for key in from_dict if key not in to_dict],
                [key for key in from_dict if key in to_dict])

    def _process(self):
        """ Figures out the operations needed to get to self.db_to

        Returns a list of operations that will migrate db_from to db_to

        For performance reasons the migrations are done with foreign key
        checks off: the first operation turns them off, and the last operation
        turns them on.

        :return: Operations needed to complete the migration
        :rtype: [mygrations.formats.mysql.mygrations.operations]
        """
        operations = []
        if self._disable_fk_checks:
            operations.append(disable_checks())

        # what tables have been added/changed/removed
        db_from_tables = self.db_from.tables if self.db_from else {}
        (tables_to_add, tables_to_remove, tables_to_update) = self._differences(db_from_tables, self.db_to.tables)

        for table_name in tables_to_add:
            operations.append(self.db_to.tables[table_name].create())

        for table_name in tables_to_update:
            target_table = self.db_to.tables[table_name]
            source_table = self.db_from.tables[table_name]
            operations.extend(source_table.to(target_table))

        for table_name in tables_to_remove:
            operations.append(remove_table(table_name))

        if self._disable_fk_checks:
            operations.append(enable_checks())

        return operations

    def _old_process(self):
        """ Figures out the operations (and proper order) need to get to self.db_to

        This particular method is no longer used but is kept around for future
        reference.  This was my original attempt, the goal of which was to make the
        system smart enough to determine the proper order of operations to avoid
        violating FK constraints.  See these:

        https://codereview.stackexchange.com/q/177780/140581
        https://softwareengineering.stackexchange.com/q/359107/243180

        However, based primarly on concerns about performance, I've decided to
        let it be dumb and simply perform all operations with FK
        checking off, and then turn it back on when done.  I'm going to leave this
        here though because I may resurrect it with a flag to enable migrations
        with FK checks on.  When I do that though it will be time to figure out
        proper topological sorting

        Excessively commented because there are a lot of details and this is a critical
        part of the process
        """

        # Our primary output is a list of operations, but there is more that we need
        # to make all of this happen.  We need a database to keep track of the
        # state of the database we are building after each operation is "applied"
        tracking_db = copy.deepcopy(self.db_from) if self.db_from else database()

        # a little bit of extra processing will simplify our algorithm by a good chunk.
        # The situation is much more complicated when we have a database we are migrating
        # from, because columns might be added/removed/changed, and it might be (for instance)
        # that the removal of a column breaks a foreign key constraint.  The general
        # ambiguities introduced by changes happening in different times/ways makes it
        # much more difficult to figure out when foreign key constraints can properly
        # be added without triggering a 1215 error.  The simplest way to straighten this
        # all out is to cheat: "mygrate" the "to" database all by itself.  Without a "from"
        # the operations are more straight-forward, and we can figure out with less effort
        # whether or not all FK constraints can be fulfilled.  If they aren't all fulfilled,
        # then just quit now before we do anything.  If they are all fulfilled then we
        # know our final table will be fine, so if we can just split off any uncertain
        # foreign key constraints and apply them all at the end when our database is done
        # being updated.  Simple(ish)!
        #if self.db_from:
        #check = mygration( self.db_to )
        #if check.errors_1215:
        #return [ check.errors_1215, [] ]

        # First figure out the status of individual tables
        db_from_tables = self.db_from.tables if self.db_from else {}
        (tables_to_add, tables_to_remove, tables_to_update) = self._differences(db_from_tables, self.db_to.tables)

        # IMPORTANT! tracking db and tables_to_add are both passed in by reference
        # (like everything in python), but in this case I actually modify them by reference.
        # not my preference, but it makes it easier here
        (errors_1215, operations) = self._process_adds(tracking_db, tables_to_add)

        # if we have errors we are done
        #if errors_1215:
        #return [ errors_1215, operations ]

        # now apply table updates.  This acts differently: it returns a dictionary with
        # two sets of operations: one to update the tables themselves, and one to update
        # the foreign keys.  The latter are applied after everything else has happened.
        fk_operations = []
        split_operations = self._process_updates(tracking_db, tables_to_update)
        # remove fks first to avoid 1215 errors caused by trying to remove a column
        # that is being used in a FK constraint that hasn't yet been removed
        if split_operations['removed_fks']:
            operations.extend(split_operations['removed_fks'])
        if split_operations['kitchen_sink']:
            operations.extend(split_operations['kitchen_sink'])
        if split_operations['fks']:
            fk_operations.extend(split_operations['fks'])

        # now that we got some tables modified let's try adding tables again
        # if we have any left.  Remember that tracking_db and tables_to_add
        # are modified in-place.  The point here is that there may be some
        # tables to add that we were not able to add before because they
        # relied on adding a column to a table before a foreign key could
        # be supported.
        if tables_to_add:
            (errors_1215, more_operations) = self._process_adds(tracking_db, tables_to_add)
            if more_operations:
                operations = operations.extend(more_operations)
            if errors_1215:
                if fk_operations:
                    operations.extend(fk_operations)
                return operations

        # At this point in time if we still have tables to add it is because
        # they have mutually-dependent foreign key constraints.  The way to
        # fix that is to be a bit smarter (but not too smart) and remove
        # from the tables all foreign key constraints that can't be added yet.
        # Then run the CREATE TABLE operations, and add the foreign key
        # constraints afterward
        for table_to_add in tables_to_add:
            new_table = self.db_to.tables[table_to_add]
            bad_constraints = tracking_db.unfulfilled_fks(new_table)
            new_table_copy = copy.deepcopy(new_table)
            create_fks = alter_table(table_to_add)
            for constraint in bad_constraints.values():
                create_fks.add_operation(add_constraint(constraint['foreign_key']))
                new_table_copy.remove_constraint(constraint['foreign_key'])
            operations.append(new_table_copy.create())
            fk_operations.append(create_fks)

        # process any remaining foreign key constraints
        if fk_operations:
            operations.extend(fk_operations)

        # finally remove any tables
        for table_to_remove in tables_to_remove:
            operations.append(remove_table(table_to_remove))
            tracking_db.remove_table(table_to_remove)

        # all done!!!
        return operations

    def _process_adds(self, tracking_db, tables_to_add):
        """ Runs through tables_to_add and resolves FK constraints to determine order to add tables in

        tracking_db and tables_to_add are passed in by reference and modified

        :returns: A list of 1215 error messages and a list of mygration operations
        :rtype: ( [{'error': string, 'foreign_key': mygrations.formats.mysql.definitions.constraint}], [mygrations.formats.mysql.mygrations.operations.operation] )
        """
        errors_1215 = []
        operations = []
        good_tables = {}

        # keep looping through tables as long as we find some to process
        # the while loop will stop under two conditions: if all tables
        # are processed or if we stop adding tables, which happens if we
        # have tables with mutualy-dependent foreign key constraints
        last_number_to_add = 0
        while tables_to_add and len(tables_to_add) != last_number_to_add:
            last_number_to_add = len(tables_to_add)
            for new_table_name in tables_to_add:
                new_table = self.db_to.tables[new_table_name]
                bad_constraints = tracking_db.unfulfilled_fks(new_table)

                # if we found no problems then we can add this table to our
                # tracking db and add the "CREATE TABLE" operation to our list of operations
                if not bad_constraints:
                    tables_to_add.remove(new_table_name)
                    operations.append(new_table.create())
                    tracking_db.add_table(new_table)
                    continue

                # the next question is whether this is a valid constraint
                # that simply can't be added yet (because it has dependencies
                # that have not been added) or if there is an actual problem
                # with the constraint.
                if new_table_name in good_tables:
                    continue

                # If we are here we have to decide if this table is fulfillable
                # eventually, or if there is a mistake with a foreign key that
                # we can't fix.  To tell the difference we just check if the
                # database we are migrating to can fulfill these foreign keys.
                broken_constraints = self.db_to.unfulfilled_fks(new_table)
                if not broken_constraints:
                    good_tables[new_table_name] = True
                    continue

                # otherwise it is no good: record as such
                tables_to_add.remove(new_table_name)
                for error in broken_constraints.values():
                    errors_1215.append(error['error'])

        return (errors_1215, operations)

    def _process_updates(self, tracking_db, tables_to_update):
        """ Runs through tables_to_update and resolves FK constraints to determine order to add them in

        tracking_db is passed in by reference and modified

        This doesn't return a list of 1215 errors because those would have been
        Taken care of the first run through when the "to" database was mygrated
        by itself.  Instead, this separates alters and addition/modification of
        foreign key updates into different operations so the foreign key updates
        can be processed separately.

        :returns: a dict
        :rtype: {'fks': list, 'kitchen_sink': list}
        """

        tables_to_update = tables_to_update[:]

        operations = {'removed_fks': [], 'fks': [], 'kitchen_sink': []}

        for update_table_name in tables_to_update:
            target_table = self.db_to.tables[update_table_name]
            source_table = self.db_from.tables[update_table_name]

            more_operations = source_table.to(target_table, True)
            if 'removed_fks' in more_operations:
                operations['removed_fks'].append(more_operations['removed_fks'])
                for operation in more_operations['removed_fks']:
                    tracking_db.apply_operation(update_table_name, operation)
            if 'fks' in more_operations:
                operations['fks'].append(more_operations['fks'])
                for operation in more_operations['fks']:
                    tracking_db.apply_operation(update_table_name, operation)
            if 'kitchen_sink' in more_operations:
                operations['kitchen_sink'].append(more_operations['kitchen_sink'])
                for operation in more_operations['kitchen_sink']:
                    tracking_db.apply_operation(update_table_name, operation)

        return operations



class rule_base(object):
    """ rule = rule_base( parser, rule, next_rule )

    Base class for rule parsers: the smallest part of our rule based parsing.  Each rule parser
    (currently rule_children, rule_delimited, rule_literal, and rule_regexp) specifies a different
    *kind* of rule to match against while parsing SQL.  Quick summary:

    ================  =======================
    Class             Description
    ================  =======================
    rule_children     Matches any number of parsers against the string
    rule_delimited    Finds delimited text values
    rule_literal      Finds literal text values
    rule_regexp       Matches a regular expression against the string
    ================  =======================

    """

    require_name = True
    require_value = True
    require_next = False

    parser_class = ''
    rule = {}
    next_rule = {}
    name = ''
    result = ''
    leftovers = ''

    def __init__(self, parser, rule, next_rule):
        """ rule = rule_base( parser, rule, next_rule )

        rule constructor.  Pass in the parser object that the rule is for, the rule
        that is being parsed, and the next rule in the parsing chain.

        rules only match at the very beginning of the string being matched.

        :param parser: The parser that this rule belongs to
        :param rule: A dictionary representing the rule configuration (see each rule class for details)
        :param next_rule: A dictionary representing the next rule
        :type parser: parser
        :type rule: dict
        :type next_rule: dict
        :returns: rule object
        """

        # we keep parser class around just for reference and better errors
        self.parser_class = parser.__class__

        # store rule and next_rule
        self.rule = rule
        self.next_rule = next_rule

        self.name = self.rule['name'] if 'name' in self.rule else ''

        # name is requied more often than not
        if self.require_name and not self.name:
            raise ValueError("name required for rule %s in class %s" % (self.rule, self.parser_class))

        # ditto rule
        if self.require_value:
            if not 'value' in self.rule or not self.rule['value']:
                raise ValueError('missing value in rule %s for class %s' % (self.rule, self.parser_class))

    def parse(self, string):
        """ rule.parse( string )

        Parse the string according to the rule class and configuration.  Returns True/False to denote
        a match, and populates self.result and self.leftovers accordingly.

        :param string: The string to parse
        :type string: str
        :returns: boolean
        """

        # why is this here?  Because python abstract classes are stange, and I don't feel like bothering with them.
        # this accomplishes the exact same thing, just at run-time instead of compile-time.  Not a big deal
        # in this case.
        raise ValueError("You forgot to extend rule_base.parse")



class remove_key:
    """ Generates a partial SQL command to drop a key from a table """

    def __init__(self, key):
        self.key = key

    def apply_to_table(self, table):
        """ Removes the key from the table

        :param table: The table to remove the key from
        :param type: mygrations.formats.mysql.definitions.table
        """
        table.remove_key(self.key)

    def __str__(self):
        if self.key.index_type == 'PRIMARY':
            return 'DROP PRIMARY KEY'
        return 'DROP KEY `%s`' % (self.key.name)



class row_update:
    """ Generates an SQL command to update a record """

    def __init__(self, table_name, data):
        if 'id' not in data:
            raise KeyError('Missing `id` column needed for update')

        if type(table_name) != str:
            self._table_name = table_name.name
        else:
            self._table_name = table_name
        self.data = data

    @property
    def table_name(self):
        """ Public getter.  Returns the name of the table.

        :returns: The table name
        :rtype: string
        """

        return self._table_name

    def __str__(self):
        updates = ', '.join([
            "`%s`='%s'" % (key, str(val).replace('\\', '\\\\')) if val is not None else ('`%s`=NULL' % key)
            for (key, val) in self.data.items() if key != 'id'
        ])
        return 'UPDATE `%s` SET %s WHERE id=%s;' % (self._table_name, updates, self.data['id'])



import unittest

from mygrations.formats.mysql.file_reader.reader import reader
class test_index_key_parsing(unittest.TestCase):
    def test_simple(self):
        parser = reader()
        returned = parser.parse(
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` VARCHAR(255) NOT NULL DEFAULT '',
                KEY `logs_message` (`message`),
                INDEX `logs_traceback` (`traceback`),
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        )

        # we should have matched
        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('logs' in parser.tables)

        self.assertTrue('logs_message' in parser.tables['logs'].indexes)
        self.assertTrue('logs_traceback' in parser.tables['logs'].indexes)
        self.assertEquals(['message'], parser.tables['logs'].indexes['logs_message'].columns)
        self.assertEquals(['traceback'], parser.tables['logs'].indexes['logs_traceback'].columns)

    def test_unique(self):
        parser = reader()
        returned = parser.parse(
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` VARCHAR(255) NOT NULL DEFAULT '',
                UNIQUE KEY `logs_message` (`message`),
                UNIQUE INDEX `logs_traceback` (`traceback`),
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        )

        # we should have matched
        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('logs' in parser.tables)

        self.assertTrue('logs_message' in parser.tables['logs'].indexes)
        self.assertTrue('logs_traceback' in parser.tables['logs'].indexes)
        self.assertEquals('UNIQUE', parser.tables['logs'].indexes['logs_message'].index_type)
        self.assertEquals('UNIQUE', parser.tables['logs'].indexes['logs_traceback'].index_type)



from mygrations.core.parse.parser import parser
class insert_values(parser):

    has_comma = False
    values = []

    rules = [{
        'type': 'literal',
        'value': '('
    }, {
        'type': 'delimited',
        'name': 'values',
        'separator': ',',
        'quote': "'"
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def __init__(self, rules=[]):

        super().__init__(rules)

        self._errors = []
        self._warnings = []
        self.values = []

    def process(self):

        self.values = [
            str(val).replace('\\\\', '\\').replace("\\'", "'") if val != 'NULL' else None
            for val in self._values['values']
        ]
        self.has_comma = True if 'ending_comma' in self._values else False



from mygrations.core.parse.parser import parser
from mygrations.formats.mysql.definitions.column import column
class type_plain(parser, column):

    has_comma = False

    # created date
    rules = [{
        'type': 'regexp',
        'value': '[^\(\s\)]+',
        'name': 'name'
    }, {
        'type': 'regexp',
        'value': '\w+',
        'name': 'type'
    }, {
        'type': 'literal',
        'value': 'NOT NULL',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'DEFAULT ([^\(\s\),]+)',
        'optional': True,
        'name': 'default'
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):

        self.has_comma = True if 'ending_comma' in self._values else False

        self._errors = []
        self._warnings = []
        self._name = self._values['name'].strip('`')
        self._length = ''
        self._column_type = self._values['type']
        self._default = self._values['default'].strip("'") if 'default' in self._values else None
        self._null = False if 'NOT NULL' in self._values else True

        # make sense of the default
        if self._default and self._default.lower() == 'null':
            self._default = None

        if self._default is None and not self._null:
            self._warnings.append(
                'Column %s is not null and has no default: you should set a default to avoid MySQL warnings' %
                (self._name)
            )



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
class test_table_difference(unittest.TestCase):
    def test_drop_columns(self):

        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `membership_id` int(10) unsigned not null,
            `task` varchar(255) DEFAULT NULL,
            `subject` text
            );
        """
        )

        # if we subtract b from a we should get some drop column queries in one alter statement
        operations = b.to(a)
        self.assertEquals(1, len(operations))
        self.assertEquals('ALTER TABLE `tasks` DROP membership_id, DROP subject;', str(operations[0]))

    def test_add_columns(self):

        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `membership_id` int(10) unsigned not null,
            `task` varchar(255) DEFAULT NULL,
            `subject` text
            );
        """
        )

        # if we subtract b from a we should get some drop column queries in one alter statement
        operations = a.to(b)
        self.assertEquals(1, len(operations))
        self.assertEquals(
            'ALTER TABLE `tasks` ADD `membership_id` INT(10) UNSIGNED NOT NULL AFTER `account_id`, ADD `subject` TEXT AFTER `task`;',
            str(operations[0])
        )

    def test_add_remove_change_columns(self):

        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) NOT NULL DEFAULT 0,
            `membership_id` int(10) unsigned not null,
            `subject` text
            );
        """
        )

        # but we can ask for it in one
        operations = a.to(b)
        self.assertEquals(1, len(operations))
        self.assertEquals(
            'ALTER TABLE `tasks` ADD `membership_id` INT(10) UNSIGNED NOT NULL AFTER `account_id`, ADD `subject` TEXT AFTER `membership_id`, CHANGE `account_id` `account_id` INT(10) NOT NULL DEFAULT 0, DROP task;',
            str(operations[0])
        )

    def test_split(self):

        a = create_parser()
        a.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) DEFAULT NULL,
            `task` varchar(255) DEFAULT NULL
            );
        """
        )

        b = create_parser()
        b.parse(
            """CREATE TABLE `tasks` (
            `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            `account_id` int(10) NOT NULL DEFAULT 0,
            `membership_id` int(10) unsigned not null,
            `subject` text,
            CONSTRAINT `tasks_account_id_ref_accounts_id` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE
            );
        """
        )

        operations = a.to(b, True)
        self.assertEquals(2, len(operations))
        self.assertEquals(
            'ALTER TABLE `tasks` ADD CONSTRAINT `tasks_account_id_ref_accounts_id` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE;',
            str(operations['fks'])
        )
        self.assertEquals(
            'ALTER TABLE `tasks` ADD `membership_id` INT(10) UNSIGNED NOT NULL AFTER `account_id`, ADD `subject` TEXT AFTER `membership_id`, CHANGE `account_id` `account_id` INT(10) NOT NULL DEFAULT 0, DROP task;',
            str(operations['kitchen_sink'])
        )



import unittest
from mygrations.formats.mysql.mygrations.operations.row_delete import row_delete
from collections import OrderedDict
class test_row_delete(unittest.TestCase):
    def test_simple(self):
        op = row_delete('a_table', 7)

        self.assertEquals("DELETE FROM `a_table` WHERE id=7;", str(op))
        self.assertEquals('a_table', op.table_name)



import unittest

from mygrations.formats.mysql.file_reader.database import database as database_reader
class test_table_1215_regressions(unittest.TestCase):
    def test_foreign_key_without_index(self):
        """ Discovered that the system was not raising an error for a foreign key that didn't have an index for the table it was attached to """

        db = database_reader([
            """CREATE TABLE `vendors` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT, PRIMARY KEY (`id`));""",
            """CREATE TABLE `payment_requests_external` (
  `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
  `account_id` int(10) UNSIGNED NOT NULL,
  `vendor_id` int(10) UNSIGNED DEFAULT NULL,
  `vendor_name` VARCHAR(255) NOT NULL DEFAULT '',
  `vendor_city` VARCHAR(255) NOT NULL DEFAULT '',
  `vendor_state` VARCHAR(255) NOT NULL DEFAULT '',
  `vendor_zip` VARCHAR(255) NOT NULL DEFAULT '',
  `vendor_list_id` VARCHAR(255) NOT NULL DEFAULT '',
  `edit_sequence` VARCHAR(255) NOT NULL DEFAULT '',
  `memo` VARCHAR(255) NOT NULL DEFAULT '',
  `request_date` INT(10) UNSIGNED NOT NULL DEFAULT 0,
  `guid` VARCHAR(255) NOT NULL DEFAULT '',
  `po_number` VARCHAR(255) NOT NULL DEFAULT '',
  `description` VARCHAR(255) NOT NULL DEFAULT '',
  `property_address1` VARCHAR(255) NOT NULL DEFAULT '',
  `property_address2` VARCHAR(255) NOT NULL DEFAULT '',
  `property_city` VARCHAR(255) NOT NULL DEFAULT '',
  `property_state` VARCHAR(255) NOT NULL DEFAULT '',
  `property_zip` VARCHAR(255) NOT NULL DEFAULT '',
  `customer_name` VARCHAR(255) NOT NULL DEFAULT '',
  `customer_list_id` VARCHAR(255) NOT NULL DEFAULT '',
  `contractor_completion_date` INT(10) UNSIGNED NOT NULL,
  `po_amount` DECIMAL(20,2) NOT NULL,
  `line_items` text,
  created INT(10) UNSIGNED NOT NULL DEFAULT 0,
  updated INT(10) UNSIGNED NOT NULL DEFAULT 0,
  PRIMARY KEY (`id`),
  KEY `account_id_pr_external` (`account_id`),
  CONSTRAINT `vendor_id_pr_external_fk` FOREIGN KEY (`vendor_id`) REFERENCES `vendors` (`id`) ON DELETE SET NULL ON UPDATE CASCADE
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci;"""
        ])

        self.assertEquals(1, len(db.errors_1215))
        self.assertEquals(
            'MySQL 1215 error for foreign key `vendor_id_pr_external_fk`: missing index. `payment_requests_external`.`vendor_id` does not have an index and therefore cannot be used in a foreign key constraint',
            db.errors_1215[0]
        )



import unittest

from mygrations.formats.mysql.file_reader.create_parser import create_parser
class test_table_text_false_positives(unittest.TestCase):
    def test_row_false_positives(self):
        """ Complicated false positives for changes on text columns

        Even if you don't specify a COLLATE or CHARACTER SET on a text column, MySQL
        may still return a COLLATE or CHARACTER SET in the `SHOW CREATE TABLE` command

        You can verify that by running these two queries:

            >>> CREATE TABLE `collate_false_positive` (
            >>>   `quickbooks_log_id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            >>>   `quickbooks_ticket_id` int(10) unsigned DEFAULT NULL,
            >>>   `batch` int(10) unsigned NOT NULL,
            >>>   `msg` text NOT NULL,
            >>>   `log_datetime` datetime NOT NULL,
            >>>   PRIMARY KEY (`quickbooks_log_id`),
            >>>   KEY `quickbooks_ticket_id` (`quickbooks_ticket_id`),
            >>>   KEY `batch` (`batch`)
            >>> ) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci;
            >>>
            >>> SHOW CREATE TABLE collate_false_positive;

        The output of the CREATE TABLE will include: `msg text COLLATE utf8_unicode_ci NOT NULL`
        which is not found in the original create table command.  If we assume (which we do)
        that any difference between the input and the output (from `SHOW CREATE TABLE`) means
        that the table needs to be altered, then this will result in us attempting to
        alter this table every time we migrate, with no affect.  This test makes sure
        we don't all into this false-positive trap.
        """
        from_table = create_parser()
        from_table.parse(
            """CREATE TABLE `collate_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `msg` text COLLATE utf8_unicode_ci NOT NULL,
  PRIMARY KEY (`id`),
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci;
        """
        )

        to_table = create_parser()
        to_table.parse(
            """CREATE TABLE `collate_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `msg` text NOT NULL,
  PRIMARY KEY (`id`),
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci;
        """
        )

        self.assertEquals(0, len(from_table.to(to_table)))

    def test_false_positive_didnt_break_real_positives(self):
        """ Make sure that the above false-positive correction didn't break real-positive detections """

        from_table = create_parser()
        from_table.parse(
            """CREATE TABLE `collate_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `msg` text COLLATE utf8_unicode_ci NOT NULL,
  PRIMARY KEY (`id`),
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci;
        """
        )

        to_table = create_parser()
        to_table.parse(
            """CREATE TABLE `collate_false_positive` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
  `msg` text COLLATE latin2_general_ci NOT NULL,
  PRIMARY KEY (`id`),
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci;
        """
        )

        ops = [str(val) for val in from_table.to(to_table)]
        self.assertEquals(
            "ALTER TABLE `collate_false_positive` CHANGE `msg` `msg` TEXT NOT NULL COLLATE 'LATIN2_GENERAL_CI';", ops[0]
        )

    def test_table_false_positives(self):
        """ This same basic bug happens at the table-level in a different way

        Even if you don't specify a COLLATE or CHARACTER SET on the table, MySQL
        may still return a COLLATE or CHARACTER SET in the `SHOW CREATE TABLE` command,
        with the value set by the default for the database.

        You can verify that by running these two queries:

            >>> CREATE TABLE `collate_false_positive` (
            >>>   `quickbooks_log_id` int(10) unsigned NOT NULL AUTO_INCREMENT,
            >>>   `quickbooks_ticket_id` int(10) unsigned DEFAULT NULL,
            >>>   `batch` int(10) unsigned NOT NULL,
            >>>   `msg` text NOT NULL,
            >>>   `log_datetime` datetime NOT NULL,
            >>>   PRIMARY KEY (`quickbooks_log_id`),
            >>>   KEY `quickbooks_ticket_id` (`quickbooks_ticket_id`),
            >>>   KEY `batch` (`batch`)
            >>> ) ENGINE=InnoDB;
            >>>
            >>> SHOW CREATE TABLE collate_false_positive;

        The output of the CREATE TABLE will include: `ENGINE=InnoDB DEFAULT CHARSET=latin1;`
        which is not found in the original create table command.  If we assume (which we do)
        that any difference between the input and the output (from `SHOW CREATE TABLE`) means
        that the table needs to be altered, then this will result in us attempting to
        alter this table every time we migrate, with no affect.  This test makes sure
        we don't all into this false-positive trap.
        """
        pass
